/*
 * Copyright (c) 2019 Samsung Electronics Co., Ltd All Rights Reserved
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#ifndef __LAUNCHPAD_HYDRA_H__
#define __LAUNCHPAD_HYDRA_H__

#ifdef __cplusplus
extern "C" {
#endif

/**
 * @brief Called when the hydra loader starts.
 * @details The callback function is called before the main loop of the hydra loader starts.
 *          In this callback, you can initialize resources like preloading libraries.
 * @since_tizen 5.5
 *
 * @param[in]   user_data       The user data passed from the callback registration function
 *
 * @see launchpad_hydra_main()
 * @see #hydra_lifecycle_callback_s
 */
typedef void (*hydra_precreate_cb)(void *user_data);

/**
 * @brief Called when the hydra loader starts.
 * @details The callback function is called before the main loop of the hydra loader starts.
 * @since_tizen 5.5
 *
 * @param[in]   user_data       The user data passed from the callback registration function
 *
 * @see launchpad_hydra_main()
 * @see #hydra_lifecycle_callback_s
 */
typedef void (*hydra_create_cb)(void *user_data);

/**
 * @brief Called when a child process of the hydra loader is created.
 * @details This callback function is called in the child process when the child process is created.
 *          After this callback function returns, the process will be terminated.
 * @since_tizen 5.5
 *
 * @param[in]   argc            The argument count
 * @param[in]   argv            The argument vector
 * @param[in]   user_data       The user data passed from the callback registration function
 * @return      @c 0 on success,
 *              otherwise a negative error value
 *
 * @see launchpad_hydra_main()
 * @see #hydra_lifecycle_callback_s
 */
typedef int (*hydra_fork_cb)(int argc, char **argv, void *user_data);

/**
 * @brief Called when the hydra loader main loop exits.
 * @details You should release the hydra loader's resources in this function.
 * @since_tizen 5.5
 *
 * @param[in]   user_data       The user data passed from the callback registration function
 * @return      @c 0 on success,
 *              otherwise a negative error value
 *
 * @see launchpad_hydra_main()
 * @see #hydra_lifecycle_callback_s
 */
typedef int (*hydra_terminate_cb)(void *user_data);

/**
 * @brief The structure type containing the set of callback functions for handling hydra lifecycle events.
 * @details It is one of the input parameters of the launchpad_hydra_main() function.
 * @since_tizen 5.5
 * @see hydra_precreate_cb()
 * @see hydra_create_cb()
 * @see hydra_fork_cb()
 * @see hydra_terminate_cb()
 */
typedef struct {
	hydra_precreate_cb precreate; /**< This callback function is called at the start of the hydra loader. */
	hydra_create_cb create; /**< This callback function is called at the start of the hydra loader. */
	hydra_fork_cb fork; /**< This callback function is called at the start of the child process. */
	hydra_terminate_cb terminate; /**< This callback function is called once after the main loop of the hydra loader exits. */
} hydra_lifecycle_callback_s;

/**
 * @brief Runs the hydra loaders' main loop until launchpad_hydra_exit() is called.
 * @since_tizen 5.5
 *
 * @param[in]   argc            The argument count
 * @param[in]   argv            The argument vector
 * @param[in]   hydra_callbacks The set of callback functions to handle hydra lifecycle events.
 * @param[in]   user_data       The user data to be passed to the callback functions
 * @return      @c 0 on success,
 *              otherwise a negative error value
 *
 * @see hydra_precreate_cb()
 * @see hydra_create_cb()
 * @see hydra_fork_cb()
 * @see hydra_terminate_cb()
 * @see launchpad_hydra_exit()
 * @see #hydra_lifecycle_callback_s
 */
int launchpad_hydra_main(int argc, char **argv,
		hydra_lifecycle_callback_s *hydra_callbacks,
		void *user_data);

/**
 * @brief Exits the main loop of hydra.
 * @details The main loop of hydra stops and hydra_terminate_cb() is called.
 * @since_tizen 5.5
 * @see launchpad_hydra_main()
 * @see hydra_terminate_cb()
 */
void launchpad_hydra_exit(void);

#ifdef __cplusplus
}
#endif

#endif  /* __LAUNCHPAD_HYDRA_H__ */
