/*
 * Copyright (c) 2016 Samsung Electronics Co., Ltd All Rights Reserved
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an AS IS BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
#define _GNU_SOURCE

#include <stdio.h>
#include <malloc.h>
#include <stdlib.h>
#include <dirent.h>
#include <string.h>
#include <unistd.h>

#include "launchpad_common.h"
#include "debugger_info.h"

#define TAG_DEBUGGER		"[DEBUGGER]"
#define TAG_NAME		"NAME"
#define TAG_EXE			"EXE"
#define TAG_APP_TYPE		"APP_TYPE"
#define TAG_EXTRA_KEY		"EXTRA_KEY"
#define TAG_EXTRA_ENV		"EXTRA_ENV"
#define TAG_UNLINK		"UNLINK"
#define TAG_ATTACH		"ATTACH"
#define TAG_LAST_EXTRA_KEY	"LAST_EXTRA_KEY"
#define TAG_DEFAULT_OPT		"DEFAULT_OPT"

struct debugger_info_s {
	char *name;
	char *exe;
	GList *app_types;
	GList *extra_key_list;
	GList *extra_env_list;
	GList *unlink_list;
	char *attach;
	GList *last_extra_key_list;
	GList *default_opt_list;
};

static struct debugger_info_s *__create_debugger_info(void)
{
	struct debugger_info_s *info;

	info = calloc(1, sizeof(struct debugger_info_s));
	if (info == NULL) {
		_E("out of memory");
		return NULL;
	}

	return info;
}

static void __destroy_debugger_info(gpointer data)
{
	struct debugger_info_s *info = (struct debugger_info_s *)data;

	if (info == NULL)
		return;

	if (info->default_opt_list)
		g_list_free_full(info->default_opt_list, free);
	if (info->last_extra_key_list)
		g_list_free_full(info->last_extra_key_list, free);
	if (info->attach)
		free(info->attach);
	if (info->unlink_list)
		g_list_free_full(info->unlink_list, free);
	if (info->extra_env_list)
		g_list_free_full(info->extra_env_list, free);
	if (info->extra_key_list)
		g_list_free_full(info->extra_key_list, free);
	if (info->app_types)
		g_list_free_full(info->app_types, free);
	if (info->exe)
		free(info->exe);
	if (info->name)
		free(info->name);
	free(info);
}

static void __parse_app_types(struct debugger_info_s *info, char *line)
{
	char *token;
	char *saveptr = NULL;

	token = strtok_r(line, " |\t\r\n", &saveptr);
	while (token) {
		info->app_types = g_list_append(info->app_types, strdup(token));
		token = strtok_r(NULL, " |\t\r\n", &saveptr);
	}
}

static GList *__parse_file(GList *list, const char *path)
{
	FILE *fp;
	char buf[LINE_MAX];
	char *tok1 = NULL;
	char *tok2 = NULL;
	struct debugger_info_s *info = NULL;

	fp = fopen(path, "rt");
	if (fp == NULL)
		return list;

	while (fgets(buf, sizeof(buf), fp) != NULL) {
		FREE_AND_NULL(tok1);
		FREE_AND_NULL(tok2);
		sscanf(buf, "%ms %ms", &tok1, &tok2);
		if (tok1 && strcasecmp(TAG_DEBUGGER, tok1) == 0) {
			if (info) {
				_D("name: %s, exe: %s", info->name, info->exe);
				list = g_list_append(list, info);
			}

			info = __create_debugger_info();
			if (info == NULL)
				break;

			continue;
		}

		if (!tok1 || !tok2)
			continue;
		if (tok1[0] == '\0' || tok2[0] == '\0' || tok1[0] == '#')
			continue;
		if (info == NULL)
			continue;

		if (strcasecmp(TAG_NAME, tok1) == 0) {
			info->name = strdup(tok2);
			if (info->name == NULL) {
				_E("out of memory");
				__destroy_debugger_info(info);
				info = NULL;
				break;
			}
		} else if (strcasecmp(TAG_EXE, tok1) == 0) {
			info->exe = strdup(tok2);
			if (info->exe == NULL) {
				_E("out of memory");
				__destroy_debugger_info(info);
				info = NULL;
				break;
			}
		} else if (strcasecmp(TAG_APP_TYPE, tok1) == 0) {
			__parse_app_types(info, &buf[strlen(tok1)]);
			if (info->app_types == NULL) {
				_E("app_types is NULL");
				__destroy_debugger_info(info);
				info = NULL;
				break;
			}
		} else if (strcasecmp(TAG_EXTRA_KEY, tok1) == 0) {
			info->extra_key_list = g_list_append(
					info->extra_key_list, strdup(tok2));
		} else if (strcasecmp(TAG_EXTRA_ENV, tok1) == 0) {
			info->extra_env_list = g_list_append(
					info->extra_env_list, strdup(tok2));
		} else if (strcasecmp(TAG_UNLINK, tok1) == 0) {
			info->unlink_list = g_list_append(info->unlink_list,
					strdup(tok2));
		} else if (strcasecmp(TAG_ATTACH, tok1) == 0) {
			info->attach = strdup(tok2);
			if (info->attach == NULL) {
				_E("attach is NULL");
				__destroy_debugger_info(info);
				info = NULL;
				break;
			}
		} else if (strcasecmp(TAG_LAST_EXTRA_KEY, tok1) == 0) {
			info->last_extra_key_list = g_list_append(
					info->last_extra_key_list,
					strdup(tok2));
		} else if (strcasecmp(TAG_DEFAULT_OPT, tok1) == 0) {
			info->default_opt_list = g_list_append(
					info->default_opt_list,
					strdup(tok2));
		}
	}
	fclose(fp);

	if (info) {
		_D("name: %s, exe: %s", info->name, info->exe);
		list = g_list_append(list, info);
	}

	if (tok1)
		free(tok1);
	if (tok2)
		free(tok2);

	return list;
}

GList *_debugger_info_load(const char *path)
{
	DIR *dp;
	struct dirent *dentry = NULL;
	GList *list = NULL;
	char buf[PATH_MAX];
	char *ext;

	if (path == NULL)
		return NULL;

	dp = opendir(path);
	if (dp == NULL)
		return NULL;

	while ((dentry = readdir(dp)) != NULL) {
		if (dentry->d_name[0] == '.')
			continue;

		ext = strrchr(dentry->d_name, '.');
		if (ext && strcmp(ext, ".debugger") == 0) {
			snprintf(buf, sizeof(buf), "%s/%s",
					path, dentry->d_name);
			list = __parse_file(list, buf);
		}
	}
	closedir(dp);

	return list;
}

void _debugger_info_unload(GList *info)
{
	if (info == NULL)
		return;

	g_list_free_full(info, __destroy_debugger_info);
}

static int __comp_name(gconstpointer a, gconstpointer b)
{
	struct debugger_info_s *info = (struct debugger_info_s *)a;

	if (info == NULL || info->name == NULL || b == NULL)
		return -1;

	if (strcasecmp(info->name, b) == 0)
		return 0;

	return -1;
}

debugger_info_h _debugger_info_find(GList *info_list, const char *name)
{
	GList *list;

	if (info_list == NULL || name == NULL)
		return NULL;

	list = g_list_find_custom(info_list, name, __comp_name);
	if (list == NULL)
		return NULL;

	return (debugger_info_h)list->data;
}

const char *_debugger_info_get_exe(debugger_info_h info)
{
	if (info == NULL)
		return NULL;

	return info->exe;
}

GList *_debugger_info_get_extra_key_list(debugger_info_h info)
{
	if (info == NULL)
		return NULL;

	return info->extra_key_list;
}

GList *_debugger_info_get_extra_env_list(debugger_info_h info)
{
	if (info == NULL)
		return NULL;

	return info->extra_env_list;
}

GList *_debugger_info_get_unlink_list(debugger_info_h info)
{
	if (info == NULL)
		return NULL;

	return info->unlink_list;
}

const char *_debugger_info_get_attach(debugger_info_h info)
{
	if (info == NULL)
		return NULL;

	return info->attach;
}

GList *_debugger_info_get_last_extra_key_list(debugger_info_h info)
{
	if (info == NULL)
		return NULL;

	return info->last_extra_key_list;
}

GList *_debugger_info_get_default_opt_list(debugger_info_h info)
{
	if (info == NULL)
		return NULL;

	return info->default_opt_list;
}
