/*
 * Copyright (c) 2015 - 2016 Samsung Electronics Co., Ltd All Rights Reserved
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an AS IS BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#define _GNU_SOURCE
#include <stdio.h>
#include <stdlib.h>
#include <sys/prctl.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <sys/capability.h>
#include <sched.h>
#include <stdbool.h>
#include <malloc.h>
#include <bundle_internal.h>
#include <security-manager.h>
#include <time.h>
#include <systemd/sd-daemon.h>
#include <glib.h>
#include <linux/limits.h>
#include <ttrace.h>
#include <vconf.h>
#include <math.h>
#include <trust-anchor.h>

#include "key.h"
#include "launcher_info.h"
#include "launchpad.h"
#include "launchpad_common.h"
#include "launchpad_config.h"
#include "launchpad_debug.h"
#include "launchpad_signal.h"
#include "launchpad_types.h"
#include "loader_info.h"
#include "perf.h"

#define AUL_PR_NAME         16
#define EXEC_CANDIDATE_EXPIRED 5
#define EXEC_CANDIDATE_WAIT 1
#define DIFF(a, b) (((a) > (b)) ? (a) - (b) : (b) - (a))
#define CANDIDATE_NONE 0
#define HYDRA_NONE 0
#define PROCESS_POOL_LAUNCHPAD_SOCK ".launchpad-process-pool-sock"
#define LAUNCHPAD_LOGGER_SOCK ".launchpad-logger-sock"
#define LOADER_PATH_DEFAULT "/usr/bin/launchpad-loader"
#define LOADER_INFO_PATH	"/usr/share/aul"
#define LAUNCHER_INFO_PATH	LOADER_INFO_PATH
#define REGULAR_UID_MIN 5000
#define PAD_ERR_FAILED			-1
#define PAD_ERR_REJECTED		-2
#define PAD_ERR_INVALID_ARGUMENT	-3
#define PAD_ERR_INVALID_PATH		-4
#define CPU_CHECKER_TIMEOUT		1000
#define PI				3.14159265

enum candidate_process_state_e {
	CANDIDATE_PROCESS_STATE_RUNNING,
	CANDIDATE_PROCESS_STATE_PAUSED,
};

typedef struct {
	int type;
	bool prepared;
	int pid; /* for hydra this pid is not the pid of hydra itself */
	         /* but pid of non-hydra candidate, which was forked from hydra */
	int hydra_pid;
	int loader_id;
	int caller_pid;
	int send_fd;
	int hydra_fd;
	int last_exec_time;
	guint source;
	guint timer;
	char *loader_path;
	char *loader_extra;
	int detection_method;
	int timeout_val;
	unsigned long long cpu_total_time;
	unsigned long long cpu_idle_time;
	int threshold;
	int threshold_max;
	int threshold_min;
	int cur_event;
	bool on_boot;
	bool app_exists;
	bool touched;
	int activation_method;
	int deactivation_method;
	unsigned int ttl;
	guint live_timer;
	int state;
	bool is_hydra;
} candidate_process_context_t;

typedef struct {
	GPollFD *gpollfd;
	int type;
	int loader_id;
} loader_context_t;

typedef struct {
	GQueue *queue;
	guint timer;
	guint idle_checker;
	candidate_process_context_t *running_cpc;
} sequencer;

struct app_launch_arg {
	const char *appid;
	const char *app_path;
	appinfo_t *menu_info;
	bundle *kb;
};

struct app_arg {
	int argc;
	char **argv;
};

struct app_info {
	const char *type;
	bool exists;
};

static int __sys_hwacc;
static GList *loader_info_list;
static int user_slot_offset;
static GList *candidate_slot_list;
static app_labels_monitor *label_monitor;
static GList *launcher_info_list;
static GHashTable *__pid_table;
static int __memory_status_low;
static int __memory_status_normal;
static sequencer __sequencer;
static int MEMORY_STATUS_LOW;
static int MEMORY_STATUS_NORMAL;

static candidate_process_context_t *__add_slot(int type, int loader_id,
		int caller_pid, const char *loader_path, const char *extra,
		int detection_method, int activation_method,
		int deactivation_method, unsigned int ttl, int timeout_val,
		int threshold_max, int threshold_min, bool on_boot,
		bool app_exists, bool is_hydra);
static int __remove_slot(int type, int loader_id);
static int __add_default_slots(void);
static gboolean __handle_idle_checker(gpointer data);
static int __add_idle_checker(int detection_method, GList *cur);
static void __dispose_candidate_process(candidate_process_context_t *cpc);
static bool __is_low_memory(void);
static void __update_slot_state(candidate_process_context_t *cpc, int method);

static gboolean __handle_queuing_slots(gpointer data)
{
	candidate_process_context_t *cpc;
	unsigned long long total = 0;
	unsigned long long idle = 0;

	if (__sequencer.idle_checker > 0)
		return G_SOURCE_CONTINUE;

	if (g_queue_is_empty(__sequencer.queue)) {
		__sequencer.timer = 0;
		return G_SOURCE_REMOVE;
	}

	cpc = (candidate_process_context_t *)g_queue_pop_head(
			__sequencer.queue);
	if (!cpc) {
		_E("Critical error!");
		__sequencer.timer = 0;
		return G_SOURCE_REMOVE;;
	}

	if (!cpc->app_exists) {
		_W("The application is not installed. Type(%d)", cpc->type);
		return G_SOURCE_CONTINUE;
	}

	if (cpc->timer) {
		g_source_remove(cpc->timer);
		cpc->timer = 0;
	}

	if (cpc->pid != CANDIDATE_NONE) {
		_W("The slot(%d) is already running. pid(%d)",
				cpc->type, cpc->pid);
		return G_SOURCE_CONTINUE;
	}

	_get_cpu_idle(&total, &idle);
	cpc->cpu_idle_time = idle;
	cpc->cpu_total_time = total;

	__sequencer.idle_checker = g_timeout_add(CPU_CHECKER_TIMEOUT,
			__handle_idle_checker, cpc);
	__sequencer.running_cpc = cpc;

	_D("[__SEQUENCER__] Add idle checker. Type(%d)", cpc->type);

	return G_SOURCE_CONTINUE;
}

static bool __sequencer_slot_is_running(candidate_process_context_t *cpc)
{
	if (__sequencer.running_cpc == cpc)
		return true;

	return false;
}

static bool __sequencer_slot_exist(candidate_process_context_t *cpc)
{
	GList *found;

	found = g_queue_find(__sequencer.queue, cpc);
	if (found)
		return true;

	return false;
}

static int __sequencer_add_slot(candidate_process_context_t *cpc)
{
	if (__sequencer_slot_exist(cpc)) {
		_W("Already exists");
		return -1;
	}

	if (__sequencer_slot_is_running(cpc)) {
		_W("slot(%d) is running", cpc->type);
		return -1;
	}

	g_queue_push_tail(__sequencer.queue, cpc);

	return 0;
}

static void __sequencer_remove_slot(candidate_process_context_t *cpc)
{
	g_queue_remove(__sequencer.queue, cpc);
}

static void __sequencer_run(void)
{
	if (__sequencer.timer)
		return;

	__sequencer.timer = g_timeout_add(500, __handle_queuing_slots, NULL);
	if (!__sequencer.timer)
		_E("Failed to add sequencer timer");
}

static void __sequencer_stop(void)
{
	if (!__sequencer.timer)
		return;

	g_source_remove(__sequencer.timer);
	__sequencer.timer = 0;
}

static bool __sequencer_queue_is_empty(void)
{
	if (g_queue_is_empty(__sequencer.queue))
		return true;

	return false;
}

static int __sequencer_init(void)
{
	_D("[__SEQUENCER__] Init");

	__sequencer.queue = g_queue_new();
	if (!__sequencer.queue) {
		_E("Out of memory");
		return -1;
	}

	return 0;
}

static void __sequencer_fini(void)
{
	_D("[__SEQUENCER__] Finish");

	if (__sequencer.idle_checker > 0)
		g_source_remove(__sequencer.idle_checker);

	if (__sequencer.timer > 0)
		g_source_remove(__sequencer.timer);

	g_queue_free(__sequencer.queue);
}

static int __make_loader_id(void)
{
	static int id = PAD_LOADER_ID_DYNAMIC_BASE;

	return ++id;
}

static candidate_process_context_t *__find_slot_from_static_type(int type)
{
	candidate_process_context_t *cpc;
	GList *iter = candidate_slot_list;

	if (type == LAUNCHPAD_TYPE_DYNAMIC ||
			type == LAUNCHPAD_TYPE_UNSUPPORTED)
		return NULL;

	while (iter) {
		cpc = (candidate_process_context_t *)iter->data;
		if (type == cpc->type)
			return cpc;

		iter = g_list_next(iter);
	}

	return NULL;
}

static candidate_process_context_t *__find_slot_from_pid(int pid)
{
	candidate_process_context_t *cpc;
	GList *iter = candidate_slot_list;

	while (iter) {
		cpc = (candidate_process_context_t *)iter->data;
		if (pid == cpc->pid)
			return cpc;

		iter = g_list_next(iter);
	}

	return NULL;
}

static candidate_process_context_t *__find_hydra_slot_from_pid(int pid)
{
	candidate_process_context_t *cpc;
	GList *iter = candidate_slot_list;

	while (iter) {
		cpc = (candidate_process_context_t *)iter->data;
		if (cpc->is_hydra && pid == cpc->hydra_pid)
			return cpc;

		iter = g_list_next(iter);
	}

	return NULL;
}

static candidate_process_context_t *__find_slot_from_caller_pid(int caller_pid)
{
	candidate_process_context_t *cpc;
	GList *iter = candidate_slot_list;

	while (iter) {
		cpc = (candidate_process_context_t *)iter->data;
		if (caller_pid == cpc->caller_pid)
			return cpc;

		iter = g_list_next(iter);
	}

	return NULL;
}

static candidate_process_context_t *__find_slot_from_loader_id(int id)
{
	candidate_process_context_t *cpc;
	GList *iter = candidate_slot_list;

	while (iter) {
		cpc = (candidate_process_context_t *)iter->data;
		if (id == cpc->loader_id)
			return cpc;

		iter = g_list_next(iter);
	}

	return NULL;
}

static candidate_process_context_t *__find_slot(int type, int loader_id)
{
	if (type == LAUNCHPAD_TYPE_DYNAMIC)
		return __find_slot_from_loader_id(loader_id);

	return __find_slot_from_static_type(type);
}

static void __kill_process(int pid)
{
	char err_str[MAX_LOCAL_BUFSZ] = { 0, };

	if (kill(pid, SIGKILL) == -1) {
		_E("send SIGKILL: %s",
				strerror_r(errno, err_str, sizeof(err_str)));
	}
}

static void __refuse_candidate_process(int server_fd)
{
	int client_fd = -1;

	if (server_fd == -1) {
		_E("arguments error!");
		goto error;
	}

	client_fd = accept(server_fd, NULL, NULL);
	if (client_fd == -1) {
		_E("accept error!");
		goto error;
	}

	close(client_fd);
	_D("refuse connection!");

error:
	return;
}

static int __accept_candidate_process(int server_fd, int *out_client_fd,
		int *out_client_pid)
{
	int client_fd = -1;
	int client_pid = 0;
	int recv_ret = 0;

	if (server_fd == -1 || out_client_fd == NULL ||
			out_client_pid == NULL) {
		_E("arguments error!");
		goto error;
	}

	client_fd = accept(server_fd, NULL, NULL);
	if (client_fd == -1) {
		_E("accept error!");
		goto error;
	}

	if (_set_sock_option(client_fd, 1) < 0) {
		_E("Failed to set sock option");
		goto error;
	}

	recv_ret = recv(client_fd, &client_pid, sizeof(client_pid),
			MSG_WAITALL);
	if (recv_ret == -1) {
		_E("recv error!");
		goto error;
	}

	*out_client_fd = client_fd;
	*out_client_pid = client_pid;

	return *out_client_fd;

error:
	if (client_fd != -1)
		close(client_fd);

	return -1;
}

static int __listen_addr(struct sockaddr_un *addr)
{
	int fd = -1;
	fd = socket(AF_UNIX, SOCK_STREAM | SOCK_CLOEXEC, 0);
	if (fd < 0) {
		_E("Socket error");
		goto error;
	}

	unlink(addr->sun_path);

	_D("bind to %s", addr->sun_path);
	if (bind(fd, (struct sockaddr *)addr, sizeof(struct sockaddr_un)) < 0) {
		_E("bind error");
		goto error;
	}

	_D("listen to %s", addr->sun_path);
	if (listen(fd, MAX_PENDING_CONNECTIONS) == -1) {
		_E("listen error");
		goto error;
	}

	SECURE_LOGD("[launchpad] done, listen fd: %d", fd);
	return fd;

error:
	if (fd != -1)
		close(fd);

	return -1;
}

static int __listen_candidate_process(int type, int loader_id)
{
	struct sockaddr_un addr;

	_D("[launchpad] enter, type: %d", type);

	memset(&addr, 0x00, sizeof(struct sockaddr_un));
	addr.sun_family = AF_UNIX;
	snprintf(addr.sun_path, sizeof(addr.sun_path), "%s/daemons/%d/%s%d-%d",
			SOCKET_PATH, getuid(), LAUNCHPAD_LOADER_SOCKET_NAME,
			type, loader_id);

	return __listen_addr(&addr);
}

static int __listen_hydra_process(int type, int loader_id)
{
	struct sockaddr_un addr;

	_D("[launchpad] enter, type: %d", type);

	memset(&addr, 0x00, sizeof(struct sockaddr_un));
	addr.sun_family = AF_UNIX;
	snprintf(addr.sun_path, sizeof(addr.sun_path), "%s/daemons/%d/%s%d-%d",
			SOCKET_PATH, getuid(), HYDRA_LOADER_SOCKET_NAME,
			type, loader_id);

	return __listen_addr(&addr);
}

static int __get_loader_id(bundle *kb)
{
	const char *val;

	val = bundle_get_val(kb, AUL_K_LOADER_ID);
	if (val == NULL)
		return -1;
	_W("Requested loader id: %s", val);

	return atoi(val);
}

static int __candidate_process_real_launch(int candidate_fd, app_pkt_t *pkt)
{
	return _send_pkt_raw(candidate_fd, pkt);
}

static int __real_send(int clifd, int ret)
{
	if (clifd < 3) {
		_E("Invalid parameter. clifd(%d)", clifd);
		return -1;
	}

	if (send(clifd, &ret, sizeof(int), MSG_NOSIGNAL) < 0) {
		if (errno == EPIPE) {
			_E("send failed due to EPIPE.");
			close(clifd);
			return -1;
		}
		_E("send fail to client");
	}

	close(clifd);
	return 0;
}

static void __send_result_to_caller(int clifd, int ret, const char *app_path)
{
	_W("send result: %d", ret);

	if (clifd == -1)
		return;

	if (ret <= 1) {
		_E("launching failed");
		__real_send(clifd, ret);
		return;
	}

	if (__real_send(clifd, ret) < 0)
		__kill_process(ret);
}

static int __fork_app_process(int (*child_fn)(void *), void *arg)
{
	int pid;
	int ret;

	pid = fork();

	if (pid == -1) {
		_E("failed to fork child process");
		return -1;
	}

	if (pid == 0) {
		_W("security_manager_prepare_app_candidate ++");
		ret = security_manager_prepare_app_candidate();
		_W("security_manager_prepare_app_candidate --");
		if (ret != SECURITY_MANAGER_SUCCESS) {
			_E("failed to prepare app candidate process (%d)", ret);
			exit(1);
		}

		ret = child_fn(arg);
		_E("failed to exec app process (%d)", errno);
		exit(ret);
	}

	return pid;
}

static int __exec_loader_process(void *arg)
{
	char **argv = arg;
	char err_buf[1024];

	_send_cmd_to_amd(LAUNCHPAD_CHILD_PROCESS);
	_signal_unblock_sigchld();
	_close_all_fds();
	_setup_stdio(basename(argv[LOADER_ARG_PATH]));

	if (execv(argv[LOADER_ARG_PATH], argv) < 0) {
		_send_message_to_logger(argv[LOADER_ARG_PATH],
			"Failed to prepare candidate process. error(%d:%s)",
			errno, strerror_r(errno, err_buf, sizeof(err_buf)));
	} else {
		_D("Succeeded to prepare candidate_process");
	}

	return -1;
}

static gboolean __handle_deactivate_event(gpointer user_data)
{
	candidate_process_context_t *cpc;

	cpc = (candidate_process_context_t *)user_data;
	__update_slot_state(cpc, METHOD_TTL);
	_D("Deactivate event: type(%d)", cpc->type);

	return G_SOURCE_REMOVE;
}

static void __set_live_timer(candidate_process_context_t *cpc)
{
	if (!cpc)
		return;

	if (cpc->deactivation_method & METHOD_TTL) {
		if (cpc->live_timer == 0) {
			cpc->live_timer = g_timeout_add_seconds(cpc->ttl,
					__handle_deactivate_event, cpc);
		}
	}
}

static int __hydra_send_request(int fd, enum hydra_cmd cmd)
{
	int sent = 0;
	int size = sizeof(cmd);
	int send_ret = 0;

	while (sent != size) {
		send_ret = send(fd, (char *)&cmd + sent,
				size - sent, MSG_NOSIGNAL);
		if (send_ret == -1) {
			_E("send error! (%d)", errno);
			return -1;
		}

		sent += send_ret;
		_D("send(%d: ret: %d) : %d / %d",
			fd, send_ret, sent, size);
	}

	return 0;
}

static int __hydra_send_launch_candidate_request(int fd)
{
	SECURE_LOGD("Send launch cmd to hydra, fd: %d", fd);
	return __hydra_send_request(fd, LAUNCH_CANDIDATE);
}

static int __prepare_candidate_process(int type, int loader_id)
{
	int pid;
	char type_str[2] = {0, };
	char loader_id_str[10] = {0, };
	char argbuf[LOADER_ARG_LEN];
	char *argv[] = {NULL, NULL, NULL, NULL, NULL, NULL, NULL};
	candidate_process_context_t *cpt = __find_slot(type, loader_id);

	if (cpt == NULL)
		return -1;

	if (cpt->is_hydra && cpt->hydra_pid != HYDRA_NONE)
		return __hydra_send_launch_candidate_request(cpt->hydra_fd);

	_D("prepare candidate process / type:%d", type);
	memset(argbuf, ' ', LOADER_ARG_LEN);
	argbuf[LOADER_ARG_LEN - 1] = '\0';
	argv[LOADER_ARG_DUMMY] = argbuf;

	cpt->last_exec_time = time(NULL);

	type_str[0] = '0' + type;
	snprintf(loader_id_str, sizeof(loader_id_str), "%d", loader_id);
	argv[LOADER_ARG_PATH] = cpt->loader_path;
	argv[LOADER_ARG_TYPE] = type_str;
	argv[LOADER_ARG_ID] = loader_id_str;
	argv[LOADER_ARG_HYDRA] = cpt->is_hydra ? "1" : "0";
	argv[LOADER_ARG_EXTRA] = cpt->loader_extra;

	pid = __fork_app_process(__exec_loader_process, argv);
	if (pid == -1) {
		_E("Failed to fork candidate_process");
		return -1;
	}

	if (cpt->is_hydra) {
		cpt->hydra_pid = pid;
	} else {
		cpt->pid = pid;
		__set_live_timer(cpt);
	}

	return 0;
}

static gboolean __handle_timeout_event(gpointer user_data)
{
	candidate_process_context_t *cpc;

	cpc = (candidate_process_context_t *)user_data;
	cpc->timer = 0;

	if (cpc->pid != CANDIDATE_NONE) {
		_W("Candidate(%d) process(%d) is running", cpc->type, cpc->pid);
		return G_SOURCE_REMOVE;
	}

	__sequencer_add_slot(cpc);
	__sequencer_run();
	return G_SOURCE_REMOVE;
}

static void __set_timer(candidate_process_context_t *cpc)
{
	if (cpc == NULL || cpc->timer > 0)
		return;

	if ((cpc->detection_method & METHOD_TIMEOUT) &&
			cpc->state == CANDIDATE_PROCESS_STATE_RUNNING) {
		cpc->timer = g_timeout_add(cpc->timeout_val,
				__handle_timeout_event, cpc);
	}
}

static void __reset_slot(candidate_process_context_t *cpc)
{
	if (cpc == NULL)
		return;

	cpc->send_fd = -1;
	cpc->prepared = false;
	cpc->pid = CANDIDATE_NONE;
	cpc->source = 0;
	cpc->timer = 0;
	cpc->live_timer = 0;
}

static void __dispose_candidate_process(candidate_process_context_t *cpc)
{
	if (!cpc)
		return;

	_D("Dispose candidate process %d", cpc->type);
	if (cpc->pid > 0) {
		_D("kill process %d", cpc->pid);
		__kill_process(cpc->pid);
	}
	if (cpc->live_timer > 0)
		g_source_remove(cpc->live_timer);
	if (cpc->source > 0)
		g_source_remove(cpc->source);
	if (cpc->timer > 0)
		g_source_remove(cpc->timer);
	if (cpc->send_fd > 0)
		close(cpc->send_fd);
	__reset_slot(cpc);
}

static void __dispose_hydra_process(candidate_process_context_t *cpc)
{
	if (!cpc)
		return;

	__dispose_candidate_process(cpc);

	_D("Dispose hydra process %d", cpc->type);
	if (cpc->hydra_pid > 0) {
		_D("kill process %d", cpc->hydra_pid);
		__kill_process(cpc->hydra_pid);
		cpc->hydra_pid = HYDRA_NONE;
	}

	if (cpc->hydra_fd > 0) {
		close(cpc->hydra_fd);
		cpc->hydra_fd = -1;
	}
}

static int __send_launchpad_loader(candidate_process_context_t *cpc,
		app_pkt_t *pkt, const char *app_path, int clifd)
{
	int pid = -1;
	int ret;

	ret = _delete_sock_path(cpc->pid, getuid());
	if (ret != 0)
		return -1;

	__candidate_process_real_launch(cpc->send_fd, pkt);
	SECURE_LOGD("Request to candidate process, pid: %d, bin path: %s",
		cpc->pid, app_path);

	pid = cpc->pid;
	cpc->pid = CANDIDATE_NONE;
	__dispose_candidate_process(cpc);
	__set_timer(cpc);
	return pid;
}

static int __normal_fork_exec(int argc, char **argv, const char *app_path)
{
	char *libdir;
	char err_buf[1024];

	_D("start real fork and exec");

	libdir = _get_libdir(app_path);
	if (libdir) {
		setenv("LD_LIBRARY_PATH", libdir, 1);
		free(libdir);
	}

	_close_all_fds();

	if (execv(argv[LOADER_ARG_PATH], argv) < 0) { /* Flawfinder: ignore */
		_send_message_to_logger(argv[LOADER_ARG_PATH],
			"Failed to execute a file. error(%d:%s)",
			errno, strerror_r(errno, err_buf, sizeof(err_buf)));
		return -1;
	}
	/* never reach*/
	return 0;
}

static int __create_launcher_argv(int *argc, char ***argv, const char *app_type)
{
	int launcher_argc;
	char **launcher_argv;
	launcher_info_h launcher_info;
	const char *exe;
	const char *extra_arg;
	GList *extra_args;
	GList *iter;
	int i;

	launcher_info = _launcher_info_find(launcher_info_list, app_type);
	if (launcher_info == NULL)
		return 0;

	exe = _launcher_info_get_exe(launcher_info);
	if (exe == NULL) {
		_E("Failed to get launcher exe");
		return -1;
	}

	extra_args = _launcher_info_get_extra_args(launcher_info);
	launcher_argc = g_list_length(extra_args) + 1;
	launcher_argv = (char **)calloc(launcher_argc, sizeof(char *));
	if (launcher_argv == NULL) {
		_E("out of memory");
		return -1;
	}

	i = LOADER_ARG_PATH;
	launcher_argv[i++] = strdup(exe);

	iter = g_list_first(extra_args);
	while (iter) {
		extra_arg = (const char *)iter->data;
		if (extra_arg)
			launcher_argv[i++] = strdup(extra_arg);

		iter = g_list_next(iter);
	}

	*argc = launcher_argc;
	*argv = launcher_argv;


	return 0;
}

static void __destroy_launcher_argv(int argc, char **argv)
{
	int i;

	if (argv == NULL)
		return;

	for (i = 0; i < argc; i++)
		free(argv[i]);
	free(argv);
}

static int __create_app_argv(int *argc, char ***argv, const char *app_path,
		bundle *kb, const char *app_type)
{
	int new_argc;
	char **new_argv;
	bool attach = false;
	struct app_arg debug_arg = {0,};
	struct app_arg launcher_arg = {0,};
	struct app_arg arg = {0,};
	struct app_arg debug_extra_arg = {0,};
	int ret;
	int i;
	int c;

	ret = _debug_create_argv(&debug_arg.argc, &debug_arg.argv, &attach);
	if (ret < 0) {
		_E("Failed to create debugger argv");
		return -1;
	}

	if (attach) {
		*argc = debug_arg.argc;
		*argv = debug_arg.argv;
		return 0;
	}

	ret = _debug_create_extra_argv(&debug_extra_arg.argc,
			&debug_extra_arg.argv);
	if (ret < 0) {
		_E("Failed to create debugger extra argv");
		_debug_destroy_argv(debug_arg.argc, debug_arg.argv);
		return -1;
	}

	ret = __create_launcher_argv(&launcher_arg.argc, &launcher_arg.argv,
			app_type);
	if (ret < 0) {
		_E("Failed to create launcher argv");
		_debug_destroy_argv(debug_extra_arg.argc, debug_extra_arg.argv);
		_debug_destroy_argv(debug_arg.argc, debug_arg.argv);
		return -1;
	}

	arg.argc = bundle_export_to_argv(kb, &arg.argv);
	if (arg.argc <= 0) {
		_E("Failed to export bundle");
		__destroy_launcher_argv(launcher_arg.argc, launcher_arg.argv);
		_debug_destroy_argv(debug_extra_arg.argc, debug_extra_arg.argv);
		_debug_destroy_argv(debug_arg.argc, debug_arg.argv);
		return -1;
	}
	arg.argv[LOADER_ARG_PATH] = strdup(app_path);

	new_argc = debug_arg.argc + launcher_arg.argc + arg.argc +
		debug_extra_arg.argc;
	if (new_argc == arg.argc) {
		*argc = arg.argc;
		*argv = arg.argv;
		return 0;
	}

	new_argv = (char **)calloc(new_argc + 1, sizeof(char *));
	if (new_argv == NULL) {
		_E("out of memory");
		free(arg.argv[LOADER_ARG_PATH]);
		bundle_free_exported_argv(arg.argc, &arg.argv);
		__destroy_launcher_argv(launcher_arg.argc, launcher_arg.argv);
		_debug_destroy_argv(debug_extra_arg.argc, debug_extra_arg.argv);
		_debug_destroy_argv(debug_arg.argc, debug_arg.argv);
		return -1;
	}

	c = LOADER_ARG_PATH;
	for (i = 0; i < debug_arg.argc; i++)
		new_argv[c++] = debug_arg.argv[i];
	for (i = 0; i < launcher_arg.argc; i++)
		new_argv[c++] = launcher_arg.argv[i];
	for (i = 0; i < arg.argc; i++)
		new_argv[c++] = arg.argv[i];
	for (i = 0; i < debug_extra_arg.argc; i++)
		new_argv[c++] = debug_extra_arg.argv[i];

	*argc = new_argc;
	*argv = new_argv;

	return 0;
}

static void __real_launch(const char *app_path, bundle *kb,
		appinfo_t *menu_info)
{
	int app_argc = 0;
	char **app_argv;
	int i;
	int ret;

	if (bundle_get_val(kb, AUL_K_DEBUG) != NULL)
		putenv("TIZEN_DEBUGGING_PORT=1");

	ret = __create_app_argv(&app_argc, &app_argv, app_path,
			kb, menu_info->app_type);
	if (ret < 0) {
		_E("Failed to create app argv");
		exit(-1);
	}

	for (i = 0; i < app_argc; i++)
		SECURE_LOGD("input argument %d : %s##", i, app_argv[i]);

	PERF("setup argument done");
	__normal_fork_exec(app_argc, app_argv, app_path);
}

static int __prepare_exec(const char *appid, const char *app_path,
			appinfo_t *menu_info, bundle *kb)
{
	char *file_name;
	char process_name[AUL_PR_NAME];
	int ret;

	/* Set new session ID & new process group ID*/
	/* In linux, child can set new session ID without check permission */
	/* TODO : should be add to check permission in the kernel*/
	setsid();

	ret = _enable_external_pkg(kb, menu_info->pkgid,
			menu_info->global ? GLOBAL_USER : getuid());
	if (ret < 0)
		return PAD_ERR_FAILED;

	if (menu_info->global)
		ret = trust_anchor_launch(menu_info->pkgid, GLOBAL_USER);
	else
		ret = trust_anchor_launch(menu_info->pkgid, getuid());
	if (ret != TRUST_ANCHOR_ERROR_NONE &&
			ret != TRUST_ANCHOR_ERROR_NOT_INSTALLED) {
		_E("trust_anchor_launch() returns %d", ret);
		return PAD_ERR_REJECTED;
	}

	/* SET PRIVILEGES*/
	_W("security_manager_prepare_app ++");
	ret = security_manager_prepare_app(appid);
	_W("security_manager_prepare_app --");
	if (ret != SECURITY_MANAGER_SUCCESS)
		return PAD_ERR_REJECTED;

	_send_cmd_to_amd(APP_STARTUP_SIGNAL);
	if (bundle_get_type(kb, AUL_K_SDK) == BUNDLE_TYPE_NONE)
		_setup_stdio(basename(app_path));

	/* SET DUMPABLE - for coredump*/
	prctl(PR_SET_DUMPABLE, 1);

	/* SET PROCESS NAME*/
	if (app_path == NULL)
		return PAD_ERR_INVALID_ARGUMENT;

	file_name = strrchr(app_path, '/');
	if (file_name == NULL)
		return PAD_ERR_INVALID_PATH;

	file_name++;
	if (*file_name == '\0')
		return PAD_ERR_INVALID_PATH;

	memset(process_name, '\0', AUL_PR_NAME);
	snprintf(process_name, AUL_PR_NAME, "%s", file_name);
	prctl(PR_SET_NAME, process_name);

	/* SET ENVIROMENT*/
	_set_env(menu_info, kb);

	ret = _wait_tep_mount(kb);
	if (ret < 0)
		return PAD_ERR_FAILED;

	if (bundle_get_type(kb, AUL_K_SDK) == BUNDLE_TYPE_NONE) {
		ret = _prepare_app_socket();
		if (ret < 0)
			return PAD_ERR_FAILED;

		ret = _prepare_id_file();
		if (ret < 0)
			return PAD_ERR_FAILED;
	}

	return 0;
}

static int __exec_app_process(void *arg)
{
	struct app_launch_arg *launch_arg = arg;
	int ret;

	_print_hwc_log("%d|after calling fork(). %s",
			getpid(), launch_arg->appid);
	PERF("fork done");
	_D("lock up test log(no error) : fork done");

	if (bundle_get_type(launch_arg->kb, AUL_K_SDK) != BUNDLE_TYPE_NONE)
		_debug_prepare_debugger(launch_arg->kb);

	_signal_unblock_sigchld();

	_delete_sock_path(getpid(), getuid());

	PERF("prepare exec - first done");
	ret = __prepare_exec(launch_arg->appid, launch_arg->app_path,
			launch_arg->menu_info, launch_arg->kb);
	if (ret < 0)
		return ret;

	PERF("prepare exec - second done");
	__real_launch(launch_arg->app_path, launch_arg->kb,
			launch_arg->menu_info);

	return PAD_ERR_FAILED;
}

static int __launch_directly(const char *appid, const char *app_path, int clifd,
		bundle *kb, appinfo_t *menu_info,
		candidate_process_context_t *cpc)
{
	struct app_launch_arg arg;
	int pid;

	arg.appid = appid;
	arg.app_path = app_path;
	arg.menu_info = menu_info;
	arg.kb = kb;

	_print_hwc_log("before calling fork(). %s", appid);
	pid = __fork_app_process(__exec_app_process, &arg);
	if (pid <= 0)
		_E("failed to fork app process");

	SECURE_LOGD("==> real launch pid : %d %s", pid, app_path);

	return pid;
}

static int __create_sock_activation(void)
{
	int fds;

	fds = sd_listen_fds(0);
	if (fds == 1)
		return SD_LISTEN_FDS_START;
	else if (fds > 1)
		_E("Too many file descriptors received.");
	else
		_D("There is no socket stream");

	return -1;
}

static int __launchpad_pre_init(int argc, char **argv)
{
	int fd;

	/* signal init*/
	_signal_init();

	/* create launchpad sock */
	fd = __create_sock_activation();
	if (fd < 0) {
		fd = _create_server_sock(PROCESS_POOL_LAUNCHPAD_SOCK);
		if (fd < 0) {
			_E("server sock error %d", fd);
			return -1;
		}
	}

	return fd;
}

static void __destroy_poll_data(gpointer data)
{
	free(data);
}

static gboolean __glib_check(GSource *src)
{
	GSList *fd_list;
	GPollFD *tmp;

	fd_list = src->poll_fds;
	do {
		tmp = (GPollFD *) fd_list->data;
		if ((tmp->revents & (G_IO_IN | G_IO_PRI | G_IO_HUP |
						G_IO_NVAL)))
			return TRUE;
		fd_list = fd_list->next;
	} while (fd_list);

	return FALSE;
}

static gboolean __glib_dispatch(GSource *src, GSourceFunc callback,
		gpointer data)
{
	return callback(data);
}

static gboolean __glib_prepare(GSource *src, gint *timeout)
{
	return FALSE;
}

static void __glib_finalize(GSource *src)
{
	GSList *fd_list;
	GPollFD *gpollfd;

	fd_list = src->poll_fds;
	do {
		gpollfd = (GPollFD *)fd_list->data;
		close(gpollfd->fd);
		g_free(gpollfd);

		fd_list = fd_list->next;
	} while (fd_list);
}

static GSourceFuncs funcs = {
	.prepare = __glib_prepare,
	.check = __glib_check,
	.dispatch = __glib_dispatch,
	.finalize = __glib_finalize
};

static guint __poll_fd(int fd, gushort events, GSourceFunc func, int type,
		int loader_id)
{
	int r;
	GPollFD *gpollfd;
	GSource *src;
	loader_context_t *lc;

	src = g_source_new(&funcs, sizeof(GSource));
	if (!src) {
		_E("out of memory");
		return 0;
	}

	gpollfd = (GPollFD *)g_malloc(sizeof(GPollFD));
	if (!gpollfd) {
		_E("out of memory");
		g_source_destroy(src);
		return 0;
	}

	gpollfd->events = events;
	gpollfd->fd = fd;

	lc = malloc(sizeof(loader_context_t));
	if (lc == NULL) {
		g_free(gpollfd);
		g_source_destroy(src);
		return 0;
	}

	lc->gpollfd = gpollfd;
	lc->type = type;
	lc->loader_id = loader_id;

	g_source_add_poll(src, gpollfd);
	g_source_set_callback(src, func,
			(gpointer) lc, __destroy_poll_data);
	g_source_set_priority(src, G_PRIORITY_DEFAULT);

	r = g_source_attach(src, NULL);
	if (r  == 0) {
		g_free(gpollfd);
		g_source_destroy(src);
		return 0;
	}

	return r;
}

static gboolean __handle_loader_client_event(gpointer data)
{
	loader_context_t *lc = (loader_context_t *) data;
	int type = lc->type;
	int loader_id = lc->loader_id;
	gushort revents = lc->gpollfd->revents;
	candidate_process_context_t *cpc = __find_slot(type, loader_id);

	if (cpc == NULL)
		return G_SOURCE_REMOVE;

	if (revents & (G_IO_HUP | G_IO_NVAL)) {
		SECURE_LOGE("Type %d candidate process was " \
				"(POLLHUP|POLLNVAL), pid: %d",
				cpc->type, cpc->pid);
		cpc->pid = CANDIDATE_NONE;
		__dispose_candidate_process(cpc);
		__prepare_candidate_process(cpc->type, cpc->loader_id);
		return G_SOURCE_REMOVE;
	}

	return G_SOURCE_CONTINUE;
}

static gboolean __handle_hydra_client_event(gpointer data)
{
	loader_context_t *lc = (loader_context_t *)data;
	int type = lc->type;
	int loader_id = lc->loader_id;
	gushort revents = lc->gpollfd->revents;
	candidate_process_context_t *cpc = __find_slot(type, loader_id);

	if (cpc == NULL)
		return G_SOURCE_REMOVE;

	if (revents & (G_IO_HUP | G_IO_NVAL)) {
		SECURE_LOGE("Type %d hydra process was " \
				"(POLLHUP|POLLNVAL), pid: %d",
				cpc->type, cpc->hydra_pid);
		__dispose_hydra_process(cpc);
		__prepare_candidate_process(cpc->type, cpc->loader_id);
		return G_SOURCE_REMOVE;
	}

	return G_SOURCE_CONTINUE;
}

static gboolean __handle_loader_event(gpointer data)
{
	loader_context_t *lc = (loader_context_t *) data;
	int fd = lc->gpollfd->fd;
	int type = lc->type;
	int loader_id = lc->loader_id;
	int client_fd;
	int client_pid;
	int ret;

	candidate_process_context_t *cpc = __find_slot(type, loader_id);

	if (cpc == NULL)
		return G_SOURCE_REMOVE;

	if (!cpc->prepared) {
		ret = __accept_candidate_process(fd, &client_fd, &client_pid);
		if (ret >= 0) {
			/* for hydra need to set pid to pid of non-hydra candidate, */
			/* which is connecting now */
			if (cpc->is_hydra)
				cpc->pid = client_pid;

			cpc->prepared = true;
			cpc->send_fd = client_fd;

			SECURE_LOGD("Type %d candidate process was connected," \
					" pid: %d", type, cpc->pid);
			cpc->source = __poll_fd(client_fd, G_IO_IN | G_IO_HUP,
					__handle_loader_client_event, type,
					loader_id);
			if (cpc->source == 0)
				close(client_fd);
		}
	} else {
		__refuse_candidate_process(fd);
		_E("Refused candidate process connection");
	}

	return G_SOURCE_CONTINUE;
}

static gboolean __handle_hydra_event(gpointer data)
{
	loader_context_t *lc = (loader_context_t *) data;
	int fd = lc->gpollfd->fd;
	int type = lc->type;
	int loader_id = lc->loader_id;
	int client_fd;
	int client_pid;
	int ret;

	candidate_process_context_t *cpc = __find_slot(type, loader_id);

	if (cpc == NULL)
		return G_SOURCE_REMOVE;

	if (!cpc->prepared) {
		ret = __accept_candidate_process(fd, &client_fd, &client_pid);
		if (ret >= 0) {
			cpc->hydra_fd = client_fd;

			SECURE_LOGD("Type %d hydra process was connected," \
					" pid: %d", type, cpc->hydra_pid);

			cpc->source = __poll_fd(client_fd, G_IO_IN | G_IO_HUP,
					__handle_hydra_client_event, type,
					loader_id);
			if (cpc->source == 0)
				close(client_fd);
		}
	} else {
		__refuse_candidate_process(fd);
		_E("Refused hydra process connection");
	}

	return G_SOURCE_CONTINUE;
}

static gboolean __handle_sigchild(gpointer data)
{
	candidate_process_context_t *cpc;
	loader_context_t *lc = (loader_context_t *) data;
	int fd = lc->gpollfd->fd;
	struct signalfd_siginfo siginfo;
	ssize_t s;
	char *appid;
	int pid;

	do {
		s = read(fd, &siginfo, sizeof(struct signalfd_siginfo));
		if (s == 0)
			break;

		if (s != sizeof(struct signalfd_siginfo))
			break;

		_signal_process_sigchld(&siginfo);

		pid = siginfo.ssi_code != SI_QUEUE ? siginfo.ssi_pid : siginfo.ssi_int;

		appid = g_hash_table_lookup(__pid_table,
				GINT_TO_POINTER(pid));
		if (appid) {
			security_manager_cleanup_app(appid, siginfo.ssi_uid, pid);
			g_hash_table_remove(__pid_table,
					GINT_TO_POINTER(pid));
		}

		cpc = __find_slot_from_pid(pid);
		if (cpc != NULL) {
			cpc->pid = CANDIDATE_NONE;
			__dispose_candidate_process(cpc);
			__prepare_candidate_process(cpc->type, cpc->loader_id);
		} else {
			cpc = __find_hydra_slot_from_pid(pid);
			if (cpc != NULL) {
				cpc->hydra_pid = HYDRA_NONE;
				__dispose_hydra_process(cpc);
				__prepare_candidate_process(cpc->type, cpc->loader_id);
			}
		}

		cpc = __find_slot_from_caller_pid(pid);
		while (cpc) {
			__remove_slot(LAUNCHPAD_TYPE_DYNAMIC, cpc->loader_id);
			cpc = __find_slot_from_caller_pid(pid);
		}
	} while (s > 0);

	return G_SOURCE_CONTINUE;
}

static gboolean __handle_label_monitor(gpointer data)
{
	candidate_process_context_t *cpc;
	GList *iter = candidate_slot_list;

	_D("__handle_label_monitor()");
	security_manager_app_labels_monitor_process(label_monitor);

	while (iter) {
		cpc = (candidate_process_context_t *)iter->data;
		if (cpc->is_hydra) {
			if (cpc->hydra_pid > 0) {
				__dispose_hydra_process(cpc);
				__prepare_candidate_process(cpc->type,
						cpc->loader_id);
			}
		} else if (cpc->pid > 0) {
			__dispose_candidate_process(cpc);
			__prepare_candidate_process(cpc->type, cpc->loader_id);
		}

		iter = g_list_next(iter);
	}

	return G_SOURCE_CONTINUE;
}

static float __interpolator(float input, int cpu_max, int cpu_min)
{
	float ret;
	float min = cpu_min / 100.0f;
	float max = cpu_max / 100.0f;

	if (input > 1.0f)
		input = 1.0f;
	if (input < 0.0f)
		input = 0.0f;

	ret = cos(input * PI) / 2.0f + 0.5f;
	ret *= max - min;
	ret += min;

	return ret;
}

static void __update_threshold(candidate_process_context_t *cpc, float delta)
{
	static float pos = 0.0f;

	pos += delta;
	if (pos < 0.0f)
		pos = 0.0f;

	if (pos > 1.0f)
		pos = 1.0f;

	cpc->threshold = (int)(__interpolator(pos,
				cpc->threshold_max, cpc->threshold_min) * 100);
	_D("[CPU] type:%d / delta:%f / input cursor : %f / threshold : %d",
			cpc->type, delta, pos, cpc->threshold);
}

static gboolean __handle_idle_checker(gpointer data)
{
	unsigned long long total = 0;
	unsigned long long idle = 0;
	int per;
	candidate_process_context_t *cpc;

	if (!data) {
		_E("Critical error!");
		__sequencer.idle_checker = 0;
		__sequencer.running_cpc = NULL;
		return G_SOURCE_REMOVE;
	}

	cpc = (candidate_process_context_t *)data;
	if (!cpc->app_exists) {
		_W("The application is not installed. Type(%d)", cpc->type);
		__sequencer.idle_checker = 0;
		__sequencer.running_cpc = NULL;
		return G_SOURCE_REMOVE;
	}

	if (cpc->state != CANDIDATE_PROCESS_STATE_RUNNING) {
		_W("Slot state is not running. Type(%d)", cpc->type);
		__sequencer.idle_checker = 0;
		__sequencer.running_cpc = NULL;
		return G_SOURCE_REMOVE;
	}

	_get_cpu_idle(&total, &idle);
	if (total == cpc->cpu_total_time)
		total++;

	per = (idle - cpc->cpu_idle_time) * 100 / (total - cpc->cpu_total_time);
	_D("[CPU] Idle : %d / type : %d", per, cpc->type);

	if (per >= cpc->threshold) {
		__update_threshold(cpc, -0.02f * (per - cpc->threshold));
		__prepare_candidate_process(cpc->type, cpc->loader_id);
		cpc->touched = true;
		__sequencer.idle_checker = 0;
		__sequencer.running_cpc = NULL;
		return G_SOURCE_REMOVE;
	}

	cpc->cpu_idle_time = idle;
	cpc->cpu_total_time = total;
	__update_threshold(cpc, 0.05f);

	return G_SOURCE_CONTINUE;
}

static int __add_idle_checker(int detection_method, GList *cur)
{
	candidate_process_context_t *cpc;
	GList *iter = cur;

	while (iter) {
		cpc = (candidate_process_context_t *)iter->data;
		if (cpc->state != CANDIDATE_PROCESS_STATE_RUNNING) {
			iter = g_list_next(iter);
			continue;
		}

		if (strcmp("null", cpc->loader_path) == 0) {
			iter = g_list_next(iter);
			continue;
		}

		if (!cpc->touched && !cpc->on_boot) {
			iter = g_list_next(iter);
			continue;
		}

		if (!cpc->app_exists) {
			iter = g_list_next(iter);
			continue;
		}

		if (cpc->pid == CANDIDATE_NONE &&
				(!cpc->touched ||
				 (cpc->detection_method & detection_method))) {
			if (cpc->timer > 0) {
				g_source_remove(cpc->timer);
				cpc->timer = 0;
			}

			cpc->cur_event = detection_method;
			__sequencer_add_slot(cpc);
			__sequencer_run();
		}

		iter = g_list_next(iter);
	}

	return -1;
}

static int __dispatch_cmd_hint(bundle *kb, int detection_method)
{
	_W("cmd hint %d", detection_method);
	__add_idle_checker(detection_method, candidate_slot_list);

	return 0;
}

static int __dispatch_cmd_add_loader(bundle *kb)
{
	const char *add_slot_str = NULL;
	const char *caller_pid = NULL;
	const char *extra;
	int lid;
	candidate_process_context_t *cpc;

	_W("cmd add loader");
	add_slot_str = bundle_get_val(kb, AUL_K_LOADER_PATH);
	caller_pid = bundle_get_val(kb, AUL_K_CALLER_PID);
	extra = bundle_get_val(kb, AUL_K_LOADER_EXTRA);

	if (add_slot_str && caller_pid) {
		lid = __make_loader_id();
		cpc = __add_slot(LAUNCHPAD_TYPE_DYNAMIC, lid, atoi(caller_pid),
				add_slot_str, extra,
				METHOD_TIMEOUT | METHOD_VISIBILITY,
				METHOD_REQUEST | METHOD_AVAILABLE_MEMORY,
				METHOD_TTL | METHOD_OUT_OF_MEMORY,
				600,
				2000,
				DEFAULT_CPU_THRESHOLD_MAX,
				DEFAULT_CPU_THRESHOLD_MIN,
				false,
				true, 0);
		__set_timer(cpc);
		return lid;
	}

	return -1;
}

static int __dispatch_cmd_remove_loader(bundle *kb)
{
	const char *id = bundle_get_val(kb, AUL_K_LOADER_ID);
	int lid;

	_W("cmd remove loader");
	if (id) {
		lid = atoi(id);
		if (__remove_slot(LAUNCHPAD_TYPE_DYNAMIC, lid) == 0)
			return 0;
	}

	return -1;
}

static int __check_caller_by_pid(int pid)
{
	int ret;
	char buf[PATH_MAX] = { 0, };

	ret = _proc_get_attr_by_pid(pid, buf, sizeof(buf));
	if (ret < 0)
		return -1;

	if (strcmp(buf, "User") == 0 ||
			strcmp(buf, "System") == 0 ||
			strcmp(buf, "System::Privileged") == 0)
		return 0;

	return -1;
}

static bool __is_hw_acc(const char *hwacc)
{
	if (strcmp(hwacc, "USE") == 0 ||
		(strcmp(hwacc, "SYS") == 0 &&
			__sys_hwacc == SETTING_HW_ACCELERATION_ON))
		return true;

	return false;
}

static candidate_process_context_t *__find_available_slot(const char *hwacc,
		const char *app_type, const char *loader_name,
		candidate_process_context_t **org_cpc)
{
	int type;
	candidate_process_context_t *cpc;
	int *a_types;
	int len = 0;
	int i;

	if (loader_name) {
		type = _loader_info_find_type_by_loader_name(loader_info_list,
				loader_name);
	} else {
		type = _loader_info_find_type(loader_info_list,
				app_type, __is_hw_acc(hwacc));
	}
	cpc = __find_slot(type, PAD_LOADER_ID_STATIC);
	if (!cpc)
		return NULL;

	*org_cpc = cpc;

	if (cpc->prepared)
		return cpc;

	a_types = _loader_get_alternative_types(loader_info_list, type, &len);
	if (!a_types)
		return NULL;

	for (i = 0; i < len; i++) {
		cpc = __find_slot(a_types[i], PAD_LOADER_ID_STATIC);
		if (!cpc)
			continue;
		if (cpc->prepared) {
			free(a_types);
			return cpc;
		}
	}

	free(a_types);
	return NULL;
}

static void __update_slot(int type, bool app_exists)
{
	candidate_process_context_t *cpc;

	cpc = __find_slot(type, PAD_LOADER_ID_STATIC);
	if (!cpc)
		return;

	cpc->app_exists = app_exists;
	if (!cpc->app_exists) {
		if (cpc->pid > 0)
			__dispose_candidate_process(cpc);
		__sequencer_remove_slot(cpc);
		if (__sequencer_queue_is_empty())
			__sequencer_stop();
	} else {
		if (cpc->state != CANDIDATE_PROCESS_STATE_RUNNING)
			return;

		if (!cpc->touched && !cpc->on_boot)
			return;

		if (cpc->timer > 0) {
			g_source_remove(cpc->timer);
			cpc->timer = 0;
		}

		if (cpc->pid == CANDIDATE_NONE) {
			__sequencer_add_slot(cpc);
			__sequencer_run();
		}
	}
}

static void __foreach_loader_info(loader_info_t *info, void *data)
{
	struct app_info *ai = (struct app_info *)data;
	bool exist;

	exist = _loader_info_exist_app_type(info, ai->type);
	if (!exist)
		return;

	info->app_exists = ai->exists;
	__update_slot(info->type, info->app_exists);
}

static int __dispatch_cmd_update_app_type(bundle *b)
{
	int r;
	struct app_info info;
	const char *is_installed;

	info.type = bundle_get_val(b, AUL_K_APP_TYPE);
	if (!info.type)
		return -1;

	is_installed = bundle_get_val(b, AUL_K_IS_INSTALLED);
	if (is_installed && !strcmp(is_installed, "true"))
		info.exists = true;
	else
		info.exists = false;

	_I("[LAUNCHPAD] type(%s), exists(%d)", info.type, info.exists);

	r = _loader_info_foreach(loader_info_list, __foreach_loader_info,
			&info);
	if (r != 0) {
		_E("Failed to retrieve loader info");
		return -1;
	}

	return 0;
}

static void __update_slot_state(candidate_process_context_t *cpc, int method)
{
	if (method == METHOD_OUT_OF_MEMORY) {
		if ((cpc->deactivation_method & method) && __is_low_memory()) {
			cpc->state = CANDIDATE_PROCESS_STATE_PAUSED;
			__dispose_candidate_process(cpc);
		} else {
			cpc->state = CANDIDATE_PROCESS_STATE_RUNNING;
			if (!cpc->touched && !cpc->on_boot)
				return;
			if (!cpc->app_exists || cpc->pid > CANDIDATE_NONE)
				return;
			if (cpc->detection_method & METHOD_TIMEOUT)
				__set_timer(cpc);
		}
	} else if (cpc->activation_method & method) {
		__update_slot_state(cpc, METHOD_OUT_OF_MEMORY);
	} else if (cpc->deactivation_method & method) {
		cpc->state = CANDIDATE_PROCESS_STATE_PAUSED;
		__dispose_candidate_process(cpc);
	} else {
		cpc->state = CANDIDATE_PROCESS_STATE_RUNNING;
		if (!cpc->touched && !cpc->on_boot)
			return;
		if (!cpc->app_exists || cpc->pid > CANDIDATE_NONE)
			return;
		if (cpc->detection_method & METHOD_TIMEOUT)
			__set_timer(cpc);
	}
}

static gboolean __handle_launch_event(gpointer data)
{
	loader_context_t *lc = (loader_context_t *) data;
	int fd = lc->gpollfd->fd;
	bundle *kb = NULL;
	app_pkt_t *pkt = NULL;
	appinfo_t *menu_info = NULL;
	candidate_process_context_t *cpc = NULL;
	candidate_process_context_t *org_cpc = NULL;
	const char *app_path = NULL;
	int pid = -1;
	int clifd = -1;
	struct ucred cr;
	int type = -1;
	int loader_id;
	int ret;

	traceBegin(TTRACE_TAG_APPLICATION_MANAGER, "LAUNCHPAD:LAUNCH");
	pkt = _accept_recv_pkt_raw(fd, &clifd, &cr);
	if (!pkt) {
		_E("packet is NULL");
		goto end;
	}

	if (cr.uid >= REGULAR_UID_MIN) {
		if (__check_caller_by_pid(cr.pid) < 0) {
			_E("Invalid caller pid");
			goto end;
		}
	}

	kb = bundle_decode(pkt->data, pkt->len);
	if (!kb) {
		_E("bundle decode error");
		goto end;
	}

	if (bundle_get_type(kb, AUL_K_SDK) != BUNDLE_TYPE_NONE)
		_debug_init();

	switch (pkt->cmd) {
	case PAD_CMD_VISIBILITY:
		ret = __dispatch_cmd_hint(kb, METHOD_VISIBILITY);
		__real_send(clifd, ret);
		clifd = -1;
		goto end;
	case PAD_CMD_ADD_LOADER:
		ret = __dispatch_cmd_add_loader(kb);
		__real_send(clifd, ret);
		clifd = -1;
		goto end;
	case PAD_CMD_REMOVE_LOADER:
		ret = __dispatch_cmd_remove_loader(kb);
		__real_send(clifd, ret);
		clifd = -1;
		goto end;
	case PAD_CMD_MAKE_DEFAULT_SLOTS:
		ret = __add_default_slots();
		if (ret != 0)
			_E("Failed to make default slots");
		__real_send(clifd, ret);
		clifd = -1;
		goto end;
	case PAD_CMD_DEMAND:
		ret = __dispatch_cmd_hint(kb, METHOD_DEMAND);
		__real_send(clifd, ret);
		clifd = -1;
		goto end;
	case PAD_CMD_PING:
		__real_send(clifd, getpid());
		clifd = -1;
		goto end;
	case PAD_CMD_UPDATE_APP_TYPE:
		__dispatch_cmd_update_app_type(kb);
		close(clifd);
		clifd = -1;
		goto end;
	}

	INIT_PERF(kb);
	PERF("packet processing start");

	menu_info = _appinfo_create(kb);
	if (menu_info == NULL) {
		_E("such pkg no found");
		goto end;
	}

	app_path = _appinfo_get_app_path(menu_info);
	if (app_path == NULL) {
		_E("app_path is NULL");
		goto end;
	}
	if (app_path[0] != '/') {
		_E("app_path is not absolute path");
		goto end;
	}

	if (menu_info->hwacc == NULL) {
		_E("[launchpad] Failed to find H/W acceleration type");
		goto end;
	}

	SECURE_LOGD("exec : %s", menu_info->app_path);
	SECURE_LOGD("comp_type : %s", menu_info->comp_type);
	SECURE_LOGD("internal pool : %s", menu_info->internal_pool);
	SECURE_LOGD("hwacc : %s", menu_info->hwacc);
	SECURE_LOGD("app_type : %s", menu_info->app_type);
	SECURE_LOGD("pkg_type : %s", menu_info->pkg_type);

	if (menu_info->comp_type &&
			strcmp(menu_info->comp_type, "svcapp") == 0) {
		loader_id = PAD_LOADER_ID_DIRECT;
	} else if (menu_info->comp_type && menu_info->app_type &&
			strcmp(menu_info->comp_type, "widgetapp") == 0 &&
			strcmp(menu_info->app_type, "webapp") == 0) {
		loader_id = PAD_LOADER_ID_DIRECT;
	} else {
		loader_id = __get_loader_id(kb);
		if (loader_id <= PAD_LOADER_ID_STATIC) {
			cpc = __find_available_slot(menu_info->hwacc,
					menu_info->app_type,
					menu_info->loader_name, &org_cpc);
		} else {
			type = LAUNCHPAD_TYPE_DYNAMIC;
			cpc = __find_slot(type, loader_id);
			if (cpc && !cpc->prepared)
				cpc = NULL;
		}
	}

	_modify_bundle(kb, cr.pid, menu_info, pkt->cmd);
	if (menu_info->appid == NULL) {
		_E("unable to get appid from menu_info");
		goto end;
	}

	PERF("get package information & modify bundle done");

	if (loader_id == PAD_LOADER_ID_DIRECT || cpc == NULL) {
		_W("Launch directly %d %p", loader_id, cpc);
		pid = __launch_directly(menu_info->appid, app_path, clifd, kb,
				menu_info, NULL);
		if (org_cpc && org_cpc->app_exists &&
				org_cpc->pid == CANDIDATE_NONE &&
				!__sequencer_slot_exist(org_cpc)) {
			if (org_cpc->timer > 0) {
				g_source_remove(org_cpc->timer);
				org_cpc->timer = 0;
			}

			__update_slot_state(org_cpc, METHOD_REQUEST);
			__set_timer(org_cpc);
		}
	} else {
		_W("Launch %d type process", cpc->type);
		pid = __send_launchpad_loader(cpc, pkt, app_path, clifd);
	}

	__send_result_to_caller(clifd, pid, app_path);
	clifd = -1;
end:
	if (clifd != -1)
		close(clifd);

	if (pid > 0) {
		_signal_send_app_launch_signal(pid, menu_info->appid);
		g_hash_table_insert(__pid_table, GINT_TO_POINTER(pid),
				strdup(menu_info->appid));
	}

	if (menu_info != NULL)
		_appinfo_free(menu_info);

	if (kb != NULL)
		bundle_free(kb);
	if (pkt != NULL)
		free(pkt);

	traceEnd(TTRACE_TAG_APPLICATION_MANAGER);

	return G_SOURCE_CONTINUE;
}

static candidate_process_context_t *__add_slot(int type, int loader_id,
		int caller_pid, const char *loader_path,
		const char *loader_extra, int detection_method,
		int activation_method, int deactivation_method,
		unsigned int ttl, int timeout_val,
		int threshold_max, int threshold_min,
		bool on_boot, bool app_exists, bool is_hydra)
{
	candidate_process_context_t *cpc;
	int fd = -1;
	guint pollfd;

	if (__find_slot(type, loader_id) != NULL)
		return NULL;

	cpc = (candidate_process_context_t *)malloc(
			sizeof(candidate_process_context_t));
	if (cpc == NULL)
		return NULL;

	cpc->type = type;
	cpc->prepared = false;
	cpc->pid = CANDIDATE_NONE;
	cpc->hydra_pid = HYDRA_NONE;
	cpc->caller_pid = caller_pid;
	cpc->loader_id = loader_id;
	cpc->send_fd = -1;
	cpc->hydra_fd = -1;
	cpc->last_exec_time = 0;
	cpc->source = 0;
	cpc->timer = 0;
	cpc->loader_path = strdup(loader_path);
	cpc->loader_extra = loader_extra ? strdup(loader_extra) : strdup("");
	cpc->detection_method = detection_method;
	cpc->timeout_val = timeout_val;
	cpc->cpu_total_time = 0;
	cpc->cpu_idle_time = 0;
	cpc->threshold = threshold_max;
	cpc->threshold_max = threshold_max;
	cpc->threshold_min = threshold_min;
	cpc->on_boot = on_boot;
	cpc->app_exists = app_exists;
	cpc->touched = false;
	cpc->cur_event = 0;
	cpc->activation_method = activation_method;
	cpc->deactivation_method = deactivation_method;
	cpc->ttl = ttl;
	cpc->live_timer = 0;
	cpc->is_hydra = is_hydra;

	if ((cpc->deactivation_method & METHOD_OUT_OF_MEMORY) &&
			__is_low_memory())
		cpc->state = CANDIDATE_PROCESS_STATE_PAUSED;
	else
		cpc->state = CANDIDATE_PROCESS_STATE_RUNNING;

	fd = __listen_candidate_process(cpc->type, cpc->loader_id);
	if (fd == -1) {
		_E("[launchpad] Listening the socket to " \
				"the type %d candidate process failed.",
				cpc->type);
		free(cpc);
		return NULL;
	}

	pollfd = __poll_fd(fd, G_IO_IN, (GSourceFunc)__handle_loader_event,
			cpc->type, cpc->loader_id);
	if (pollfd == 0) {
		close(fd);
		free(cpc);
		return NULL;
	}

	if (is_hydra) {
		fd = __listen_hydra_process(cpc->type, cpc->loader_id);
		if (fd == -1) {
			_E("[launchpad] Listening the socket to " \
					"the type %d hydra process failed.",
					cpc->type);
			free(cpc);
			return NULL;
		}

		pollfd = __poll_fd(fd, G_IO_IN, (GSourceFunc)__handle_hydra_event,
			cpc->type, cpc->loader_id);
		if (pollfd == 0) {
			close(fd);
			free(cpc);
			return NULL;
		}
	}

	candidate_slot_list = g_list_append(candidate_slot_list, cpc);

	return cpc;
}

static int __remove_slot(int type, int loader_id)
{
	candidate_process_context_t *cpc;
	GList *iter;

	iter = candidate_slot_list;
	while (iter) {
		cpc = (candidate_process_context_t *)iter->data;
		if (type == cpc->type && loader_id == cpc->loader_id) {
			__dispose_candidate_process(cpc);
			candidate_slot_list = g_list_delete_link(
					candidate_slot_list, iter);
			free(cpc->loader_path);
			if (cpc->loader_extra)
				free(cpc->loader_extra);

			free(cpc);
			return 0;
		}

		iter = g_list_next(iter);
	}

	return -1;
}

static int __init_launchpad_fd(int argc, char **argv)
{
	int fd = -1;
	guint pollfd;

	fd = __launchpad_pre_init(argc, argv);
	if (fd < 0) {
		_E("launchpad pre init failed");
		return -1;
	}

	pollfd = __poll_fd(fd, G_IO_IN, (GSourceFunc)__handle_launch_event, 0,
			0);
	if (pollfd == 0) {
		close(fd);
		return -1;
	}

	return 0;
}

static int __init_sigchild_fd(void)
{
	int fd = -1;
	guint pollfd;

	fd = _signal_get_sigchld_fd();
	if (fd < 0) {
		_E("failed to get sigchld fd");
		return -1;
	}

	pollfd = __poll_fd(fd, G_IO_IN, (GSourceFunc)__handle_sigchild, 0, 0);
	if (pollfd == 0) {
		close(fd);
		return -1;
	}

	return 0;
}

static int __init_label_monitor_fd(void)
{
	int r;
	int fd = -1;
	guint pollfd;

	r = security_manager_app_labels_monitor_init(&label_monitor);
	if (r != SECURITY_MANAGER_SUCCESS)
		return -1;

	r = security_manager_app_labels_monitor_process(label_monitor);
	if (r != SECURITY_MANAGER_SUCCESS)
		goto err;

	security_manager_app_labels_monitor_get_fd(label_monitor, &fd);
	if (fd < 0) {
		_E("failed to get fd");
		goto err;
	}

	pollfd = __poll_fd(fd, G_IO_IN,
			(GSourceFunc)__handle_label_monitor, 0, 0);
	if (pollfd == 0)
		goto err;

	return 0;

err:
	if (fd > 0)
		close(fd);

	if (label_monitor) {
		security_manager_app_labels_monitor_finish(label_monitor);
		label_monitor = NULL;
	}

	return -1;
}

static int __verify_loader_caps(const char *loader)
{
	cap_t cap_d;
	cap_flag_value_t eff_state;
	cap_flag_value_t inh_state;
	cap_value_t values[] = {CAP_SETGID, CAP_MAC_ADMIN};
	int r;
	int i;
	int size = ARRAY_SIZE(values);

	/* If Dytransition feature is enabled, CAP_MAC_ADMIN is unnecessary */
	if (label_monitor)
		size--;

	cap_d = cap_get_file(loader);
	if (!cap_d) {
		_E("Failed to get cap from file(%s)", loader);
		return -1;
	}

	for (i = 0; i < size; i++) {
		r = cap_get_flag(cap_d, values[i], CAP_INHERITABLE, &inh_state);
		if (r != 0) {
			_E("Failed to get cap inh - errno(%d)", errno);
			cap_free(cap_d);
			return -1;
		}

		r = cap_get_flag(cap_d, values[i], CAP_EFFECTIVE, &eff_state);
		if (r != 0) {
			_E("Failed to get cap eff - errno(%d)", errno);
			cap_free(cap_d);
			return -1;
		}

		if ((inh_state != CAP_SET) || (eff_state != CAP_SET)) {
			_E("The %s doesn't have %d cap", loader, values[i]);
			cap_free(cap_d);
			return -1;
		}
	}
	cap_free(cap_d);

	return 0;
}

static void __add_slot_from_info(gpointer data, gpointer user_data)
{
	loader_info_t *info = (loader_info_t *)data;
	candidate_process_context_t *cpc;
	bundle_raw *extra = NULL;
	int len;

	if (!strcmp(info->exe, "null")) {
		cpc = __add_slot(LAUNCHPAD_TYPE_USER + user_slot_offset,
				PAD_LOADER_ID_DIRECT,
				0, info->exe, NULL,
				0, 0, 0, 0, 0,
				info->cpu_threshold_max,
				info->cpu_threshold_min,
				false,
				info->app_exists, info->is_hydra);
		if (cpc == NULL)
			return;

		info->type = LAUNCHPAD_TYPE_USER + user_slot_offset;
		user_slot_offset++;
		return;
	}

	if (access(info->exe, F_OK | X_OK) == 0) {
		if (__verify_loader_caps(info->exe) < 0)
			return;

		if (info->extra)
			bundle_encode(info->extra, &extra, &len);

		cpc = __add_slot(LAUNCHPAD_TYPE_USER + user_slot_offset,
				PAD_LOADER_ID_STATIC,
				0, info->exe, (char *)extra,
				info->detection_method,
				info->activation_method,
				info->deactivation_method,
				info->ttl,
				info->timeout_val,
				info->cpu_threshold_max,
				info->cpu_threshold_min,
				info->on_boot,
				info->app_exists, info->is_hydra);
		if (cpc == NULL)
			return;

		info->type = LAUNCHPAD_TYPE_USER + user_slot_offset;
		user_slot_offset++;
	}
}

static int __add_default_slots(void)
{
	if (loader_info_list)
		_loader_info_dispose(loader_info_list);

	loader_info_list = _loader_info_load(LOADER_INFO_PATH);
	if (loader_info_list == NULL)
		return -1;

	user_slot_offset = 0;
	g_list_foreach(loader_info_list, __add_slot_from_info, NULL);
	__add_idle_checker(0, candidate_slot_list);

	return 0;
}

static bool __is_low_memory(void)
{
	if (__memory_status_low >= MEMORY_STATUS_LOW)
		return true;
	return false;
}

static void __hw_acceleration_changed_cb(keynode_t *key, void *data)
{
	__sys_hwacc = vconf_keynode_get_int(key);
	_D("sys hwacc: %d", __sys_hwacc);
}

static void __update_lang(keynode_t *node, void *user_data)
{
	char *lang = vconf_keynode_get_str(node);

	if (lang)
		setenv("LANG", lang, 1);
}

static void __memory_status_low_changed_cb(keynode_t *node, void *data)
{
	candidate_process_context_t *cpc;
	GList *iter;

	__memory_status_low = vconf_keynode_get_int(node);
	if (__memory_status_low >= MEMORY_STATUS_LOW) {
		_W("Low memory");
		iter = candidate_slot_list;
		while (iter) {
			cpc = (candidate_process_context_t *)iter->data;
			__update_slot_state(cpc, METHOD_OUT_OF_MEMORY);
			iter = g_list_next(iter);
		}
	}
}

static void __memory_status_normal_changed_cb(keynode_t *node, void *data)
{
	candidate_process_context_t *cpc;
	GList *iter;

	__memory_status_normal = vconf_keynode_get_int(node);
	if (__memory_status_normal == MEMORY_STATUS_NORMAL) {
		_W("Normal");
		iter = candidate_slot_list;
		while (iter) {
			cpc = (candidate_process_context_t *)iter->data;
			__update_slot_state(cpc, METHOD_AVAILABLE_MEMORY);
			iter = g_list_next(iter);
		}
	}
}

static void __unregister_vconf_events(void)
{
	const char *key;
	config_type_e type;

	type = CONFIG_TYPE_MEMORY_STATUS_NORMAL_KEY;
	key = _config_get_string_value(type);
	vconf_ignore_key_changed(key, __memory_status_normal_changed_cb);

	type = CONFIG_TYPE_MEMORY_STATUS_LOW_KEY;
	key = _config_get_string_value(type);
	vconf_ignore_key_changed(key, __memory_status_low_changed_cb);

	vconf_ignore_key_changed(VCONFKEY_LANGSET,
			__update_lang);
	vconf_ignore_key_changed(VCONFKEY_SETAPPL_APP_HW_ACCELERATION,
			__hw_acceleration_changed_cb);
}

static int __register_vconf_events(void)
{
	int r;
	char *lang;
	const char *key;
	config_type_e type;

	r = vconf_get_int(VCONFKEY_SETAPPL_APP_HW_ACCELERATION, &__sys_hwacc);
	if (r != VCONF_OK)
		_E("Failed to get vconf hw acceleration. err = %d", r);

	r = vconf_notify_key_changed(VCONFKEY_SETAPPL_APP_HW_ACCELERATION,
			__hw_acceleration_changed_cb, NULL);
	if (r != VCONF_OK) {
		_E("Failed to register callback for hw acceleration. err = %d",
				r);
	}

	lang = vconf_get_str(VCONFKEY_LANGSET);
	if (lang) {
		setenv("LANG", lang, 1);
		free(lang);
	}

	r = vconf_notify_key_changed(VCONFKEY_LANGSET, __update_lang, NULL);
	if (r != VCONF_OK)
		_E("Failed to register callback for langset. err = %d", r);

	type = CONFIG_TYPE_MEMORY_STATUS_LOW_KEY;
	key = _config_get_string_value(type);
	type = CONFIG_TYPE_MEMORY_STATUS_LOW_VALUE;
	MEMORY_STATUS_LOW = _config_get_int_value(type);

	r = vconf_get_int(key, &__memory_status_low);
	if (r != VCONF_OK)
		_E("Failed to get vconf low memory. err = %d", r);

	r = vconf_notify_key_changed(key,
			__memory_status_low_changed_cb, NULL);
	if (r != 0)
		_E("Failed to register callback for low memory. err = %d", r);

	type = CONFIG_TYPE_MEMORY_STATUS_NORMAL_KEY;
	key = _config_get_string_value(type);
	type = CONFIG_TYPE_MEMORY_STATUS_NORMAL_VALUE;
	MEMORY_STATUS_NORMAL = _config_get_int_value(type);

	r = vconf_get_int(key, &__memory_status_normal);
	if (r != VCONF_OK)
		_E("Failed to get vconf normal memory. err = %d", r);

	r = vconf_notify_key_changed(key,
			__memory_status_normal_changed_cb, NULL);
	if (r != 0)
		_E("Failed to register callback for normal memory. err = %d", r);

	return 0;
}

static gboolean __handle_logger(gpointer data)
{
	loader_context_t *lc = (loader_context_t *)data;
	int fd = lc->gpollfd->fd;
	app_pkt_t *pkt = NULL;
	struct ucred cr;
	int clifd = -1;

	pkt = _accept_recv_pkt_raw(fd, &clifd, &cr);
	if (!pkt) {
		_E("Failed to receive the packet");
		return G_SOURCE_CONTINUE;
	}

	if (getuid() != cr.uid) {
		_E("Invalid caller");
		goto end;
	}

	if (pkt->len <= 0) {
		_E("Invalid message");
		goto end;
	}

	SECURE_LOGE("[%d] %s", cr.pid, (const char *)pkt->data);

end:
	if (clifd != -1)
		close(clifd);
	if (pkt)
		free(pkt);

	return G_SOURCE_CONTINUE;
}

static int __init_logger_fd(void)
{
	int fd;
	guint pollfd;

	fd = _create_server_sock(LAUNCHPAD_LOGGER_SOCK);
	if (fd < 0) {
		_E("Failed to create logger socker");
		return -1;
	}

	pollfd = __poll_fd(fd, G_IO_IN, (GSourceFunc)__handle_logger, 0, 0);
	if (pollfd == 0) {
		close(fd);
		return -1;
	}

	return 0;
}

static int __before_loop(int argc, char **argv)
{
	int ret;

	ret = __sequencer_init();
	if (ret < 0) {
		_E("Failed to initialize sequencer");
		return -1;
	}

	ret = __init_sigchild_fd();
	if (ret != 0) {
		_E("__init_sigchild_fd() failed");
		return -1;
	}

	ret = __init_launchpad_fd(argc, argv);
	if (ret != 0) {
		_E("__init_launchpad_fd() failed");
		return -1;
	}

	ret = __init_logger_fd();
	if (ret != 0) {
		_E("__init_logger_fd() failed");
		return -1;
	}

	ret = __init_label_monitor_fd();
	if (ret != 0)
		_W("Failed to initialize label monitor");

	ret = _config_init();
	if (ret != 0)
		_W("Failed to initialize config");

	__add_default_slots();
	launcher_info_list = _launcher_info_load(LAUNCHER_INFO_PATH);

	ret = _send_cmd_to_amd(LAUNCHPAD_LAUNCH_SIGNAL);
	if (ret < 0)
		_W("Failed to send cmd(%d) to amd", LAUNCHPAD_LAUNCH_SIGNAL);

	__pid_table = g_hash_table_new_full(g_direct_hash, g_direct_equal,
			NULL, free);
	if (!__pid_table) {
		_E("Failed to create pid table");
		return -1;
	}

	__register_vconf_events();

	return 0;
}

static void __after_loop(void)
{
	__unregister_vconf_events();
	if (__pid_table)
		g_hash_table_destroy(__pid_table);

	if (_send_cmd_to_amd(LAUNCHPAD_DEAD_SIGNAL) < 0)
		_W("Failed to send cmd(%d) to amd", LAUNCHPAD_DEAD_SIGNAL);

	_debug_fini();
	_launcher_info_unload(launcher_info_list);
	_config_fini();

	if (label_monitor)
		security_manager_app_labels_monitor_finish(label_monitor);

	__sequencer_fini();
}

int main(int argc, char **argv)
{
	GMainLoop *mainloop = NULL;

	mainloop = g_main_loop_new(NULL, FALSE);
	if (!mainloop) {
		_E("Failed to create glib main loop");
		return -1;
	}

	if (__before_loop(argc, argv) != 0) {
		_E("process-pool Initialization failed!");
		return -1;
	}

#ifdef TIZEN_FEATURE_PRIORITY_CHANGE
	_set_priority(-12);
#endif
	g_main_loop_run(mainloop);

	__after_loop();

	return -1;
}

