/*
 * Copyright (c) 2015 - 2016 Samsung Electronics Co., Ltd All Rights Reserved
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an AS IS BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#define _GNU_SOURCE

#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/mount.h>
#include <fcntl.h>
#include <dirent.h>
#include <stdlib.h>
#include <stdio.h>
#include <sys/xattr.h>
#include <errno.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <linux/limits.h>
#include <unistd.h>
#include <sys/capability.h>
#include <tzplatform_config.h>
#include <stdio.h>
#include <stdbool.h>
#include <systemd/sd-journal.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <sys/personality.h>
#include <sys/prctl.h>
#include <dbus/dbus.h>

#include "launchpad_common.h"
#include "launchpad_types.h"
#include "key.h"

#define MAX_PATH_LEN	1024
#define MAX_CMD_BUFSZ 1024

#define MAX_PENDING_CONNECTIONS 10
#define AUL_PKT_HEADER_SIZE (sizeof(int) + sizeof(int) + sizeof(int))
#define PATH_AMD_SOCK "/run/aul/daemons/.amd-sock"
#define PATH_DEV_NULL "/dev/null"
#define MAX_TEP_IS_MOUNT_RETRY_CNT 100
#define MAX_PAYLOAD_SIZE (1024 * 1024 * 1)

#define TEP_BUS_NAME			"org.tizen.system.deviced"
#define TEP_OBJECT_PATH			"/Org/Tizen/System/DeviceD/Tzip"
#define TEP_INTERFACE_NAME		"org.tizen.system.deviced.Tzip"
#define TEP_IS_MOUNTED_METHOD		"IsMounted"

#define APP2SD_BUS_NAME			"org.tizen.app2sd"
#define APP2SD_OBJECT_PATH		"/org/tizen/app2sd"
#define APP2SD_INTERFACE_NAME		"org.tizen.app2sd"
#define APP2SD_ONDEMANDSETUPINIT_METHOD	"OndemandSetupInit"
#define APP2SD_RETRY_MAX		5
#define APP2SD_WAIT_USEC		(1000000 / 2) /* 0.5 sec */

#ifndef PR_TASK_PERF_USER_TRACE
#define PR_TASK_PERF_USER_TRACE 666
#endif

static int __read_proc(const char *path, char *buf, int size)
{
	int fd;
	int ret;

	if (buf == NULL || path == NULL)
		return -1;

	fd = open(path, O_RDONLY);
	if (fd < 0)
		return -1;

	ret = read(fd, buf, size - 1);
	if (ret <= 0) {
		close(fd);
		return -1;
	}

	buf[ret] = 0;
	close(fd);

	return ret;
}

void _get_cpu_idle(unsigned long long *total, unsigned long long *idle)
{
	FILE *fp;
	int i;
	unsigned long long sum = 0;
	unsigned long long val;
	unsigned long long iv = 0;
	char buf[4] = { 0, };

	fp = fopen("/proc/stat", "rt");

	if (fp == NULL)
		return;

	if (fscanf(fp, "%3s", buf) == -1) {
		fclose(fp);
		return;
	}

	for (i = 0; i < 10; i++) {
		if (fscanf(fp, "%lld", &val) == -1) {
			fclose(fp);
			return;
		}

		if (sum + val < sum) {
			_E("overflow");
			fclose(fp);
			return;
		}

		sum += val;
		if (i == 3) /* idle */
			iv = val;
	}

	fclose(fp);

	*total = sum;
	*idle = iv;
}

int _set_sock_option(int fd, int cli)
{
	struct timeval tv = { 5, 200 * 1000 };  /*  5.2 sec */
	int size;
	int flag;
	int ret;

	size = AUL_SOCK_MAXBUFF;
	ret = setsockopt(fd, SOL_SOCKET, SO_SNDBUF, &size, sizeof(size));
	if (ret < 0) {
		_E("Failed to set SO_SNDBUF option on socket. errno(%d)",
				errno);
		return -1;
	}

	ret = setsockopt(fd, SOL_SOCKET, SO_RCVBUF, &size, sizeof(size));
	if (ret < 0) {
		_E("Failed to set SO_RCVBUF option on socket. errno(%d)",
				errno);
		return -1;
	}

	if (cli) {
		ret = setsockopt(fd, SOL_SOCKET, SO_RCVTIMEO, &tv, sizeof(tv));
		if (ret < 0) {
			_E("Failed to set SO_RCVTIMEO option on socket. " \
					"errno(%d)", errno);
			return -1;
		}

		flag = fcntl(fd, F_GETFD);
		flag |= FD_CLOEXEC;
		ret = fcntl(fd, F_SETFD, flag);
		if (ret < 0) {
			_E("Failed to manipulate fd(F_SETFD), errno(%d)",
					errno);
			return -1;
		}
	}

	return 0;
}

static int __parse_app_path(const char *arg, char *out, int out_size)
{
	register int i;
	int state = 1;
	char *start_out = out;

	if (arg == NULL || out == NULL) {
		/* Handles null buffer*/
		return 0;
	}

	for (i = 0; out_size > 1; i++) {
		switch (state) {
		case 1:
			switch (arg[i]) {
			case ' ':
			case '\t':
				state = 5;
				break;
			case '\0':
				state = 7;
				break;
			case '\"':
				state = 2;
				break;
			case '\\':
				state = 4;
				break;
			default:
				*out = arg[i];
				out++;
				out_size--;
				break;
			}
			break;
		case 2: /* escape start*/
			switch (arg[i]) {
			case '\0':
				state = 6;
				break;
			case '\"':
				state = 1;
				break;
			default:
				*out = arg[i];
				out++;
				out_size--;
				break;
			}
			break;
		case 4: /* character escape*/
			if (arg[i] == '\0') {
				state = 6;
			} else {
				*out = arg[i];
				out++;
				out_size--;
				state = 1;
			}
			break;
		case 5: /* token*/
			if (out != start_out) {
				*out = '\0';
				out_size--;
				return i;
			}
			i--;
			state = 1;
			break;
		case 6:
			return -1;  /* error*/
		case 7: /* terminate*/
			*out = '\0';
			out_size--;
			return 0;
		default:
			state = 6;
			break;  /* error*/
		}
	}

	if (out_size == 1)
		*out = '\0';

	/* Buffer overflow*/
	return -2;
}

static int __create_client_socket(const char *path)
{
	struct sockaddr_un addr = { 0, };
	int retry = CONNECT_RETRY_COUNT;
	int fd;

	fd = socket(AF_UNIX, SOCK_STREAM | SOCK_CLOEXEC, 0);
	if (fd < 0) {
		_E("Failed to create socket(%s). errno(%d)", path, errno);
		return -1;
	}

	addr.sun_family = AF_UNIX;
	snprintf(addr.sun_path, sizeof(addr.sun_path), "%s", path);
	while (connect(fd, (struct sockaddr *)&addr, sizeof(addr)) < 0) {
		if (errno != ETIMEDOUT || retry <= 0) {
			_E("Failed to connect socket(%s). errno(%d)",
					path, errno);
			close(fd);
			return -1;
		}

		usleep(CONNECT_RETRY_TIME);
		retry--;
		_W("Retry(%d) to connect %s", retry, path);
	}

	return fd;
}

int _send_cmd_to_amd(int cmd)
{
	app_pkt_t pkt = {0,};
	int ret;
	int fd;

	fd = __create_client_socket(PATH_AMD_SOCK);
	if (fd < 0)
		return -1;

	pkt.cmd = cmd;
	ret = send(fd, &pkt, sizeof(app_pkt_t), MSG_NOSIGNAL);
	if (ret <= 0) {
		_E("Failed to send cmd(%d), errno(%d)", cmd, errno);
		close(fd);
		return -ECOMM;
	}

	close(fd);
	return 0;
}

int _create_server_sock(const char *name)
{
	struct sockaddr_un saddr;
	int fd;

	fd = socket(AF_UNIX, SOCK_STREAM | SOCK_CLOEXEC, 0);
	/*  support above version 2.6.27*/
	if (fd < 0) {
		if (errno == EINVAL) {
			fd = socket(AF_UNIX, SOCK_STREAM, 0);
			if (fd < 0) {
				_E("second chance - socket create error");
				return -1;
			}
		} else {
			_E("socket error");
			return -1;
		}
	}

	memset(&saddr, 0, sizeof(saddr));
	saddr.sun_family = AF_UNIX;

	snprintf(saddr.sun_path, sizeof(saddr.sun_path),
			"%s/daemons/%d/%s",
			SOCKET_PATH, getuid(), name);
	unlink(saddr.sun_path);

	if (bind(fd, (struct sockaddr *)&saddr, sizeof(saddr)) < 0) {
		_E("bind error");
		close(fd);
		return -1;
	}

	if (_set_sock_option(fd, 0) < 0) {
		_E("Failed to set sock option");
		close(fd);
		return -1;
	}

	if (listen(fd, 128) == -1) {
		_E("listen error");
		close(fd);
		return -1;
	}

	return fd;
}

app_pkt_t *_recv_pkt_raw(int fd)
{
	int len;
	int ret;
	unsigned char buf[AUL_SOCK_MAXBUFF];
	app_pkt_t *pkt;
	int cmd;
	int datalen;
	int opt;

retry_recv:
	/* receive header(cmd, datalen) */
	len = recv(fd, buf, AUL_PKT_HEADER_SIZE, 0);
	if (len < 0)
		if (errno == EINTR)
			goto retry_recv;

	if (len < AUL_PKT_HEADER_SIZE) {
		_E("recv error");
		return NULL;
	}
	memcpy(&cmd, buf, sizeof(int));
	memcpy(&datalen, buf + sizeof(int), sizeof(int));
	memcpy(&opt, buf + sizeof(int) + sizeof(int), sizeof(int));

	if (datalen < 0 || datalen > MAX_PAYLOAD_SIZE) {
		_E("Invalid protocol. datalen(%d)", datalen);
		return NULL;
	}

	/* allocate for a null byte */
	pkt = (app_pkt_t *)calloc(1, AUL_PKT_HEADER_SIZE + datalen + 1);
	if (pkt == NULL) {
		_E("failed to alloc app_pkt_t");
		return NULL;
	}
	pkt->cmd = cmd;
	pkt->len = datalen;
	pkt->opt = opt;

	len = 0;
	while (len != pkt->len) {
		ret = recv(fd, pkt->data + len, pkt->len - len, 0);
		if (ret < 0) {
			_E("recv error %d %d", len, pkt->len);
			free(pkt);
			return NULL;
		}
		len += ret;
		_D("recv len %d %d", len, pkt->len);
	}

	return pkt;
}

app_pkt_t *_accept_recv_pkt_raw(int fd, int *clifd, struct ucred *cr)
{
	struct sockaddr_un aul_addr = { 0, };
	int sun_size;
	app_pkt_t *pkt;
	int newfd;
	int cl = sizeof(struct ucred);

	sun_size = sizeof(struct sockaddr_un);

	newfd = accept(fd, (struct sockaddr *)&aul_addr,
			(socklen_t *) &sun_size);
	if (newfd == -1) {
		if (errno != EINTR)
			_E("accept error");
		return NULL;
	}

	if (getsockopt(newfd, SOL_SOCKET, SO_PEERCRED, cr,
			(socklen_t *) &cl) < 0) {
		_E("peer information error");
		close(newfd);
		return NULL;
	}

	if (_set_sock_option(newfd, 1) < 0) {
		_E("Failed to set sock option");
		close(newfd);
		return NULL;
	}

	pkt = _recv_pkt_raw(newfd);
	if (pkt == NULL) {
		_E("failed to receive pkt from client");
		close(newfd);
		return NULL;
	}

	*clifd = newfd;
	return pkt;
}

int _send_pkt_raw(int client_fd, app_pkt_t *pkt)
{
	int send_ret = 0;
	int pkt_size = 0;
	int sent = 0;

	if (client_fd == -1 || pkt == NULL) {
		_E("arguments error!");
		return -1;
	}

	pkt_size = AUL_PKT_HEADER_SIZE + pkt->len;

	while (sent != pkt_size) {
		send_ret = send(client_fd, (char *)pkt + sent,
				pkt_size - sent, MSG_NOSIGNAL);
		if (send_ret == -1) {
			_E("send error! (%d)", errno);
			return -1;
		}

		sent += send_ret;
		_D("send(%d: ret: %d) : %d / %d",
			client_fd, send_ret, sent, pkt_size);
	}

	return 0;
}

appinfo_t *_appinfo_create(bundle *kb)
{
	appinfo_t *menu_info;
	const char *ptr = NULL;

	menu_info = calloc(1, sizeof(appinfo_t));
	if (menu_info == NULL)
		return NULL;

	ptr = bundle_get_val(kb, AUL_K_APPID);
	if (ptr)
		menu_info->appid = strdup(ptr);
	ptr = bundle_get_val(kb, AUL_K_EXEC);
	if (ptr)
		menu_info->app_path = strdup(ptr);
	if (menu_info->app_path != NULL)
		menu_info->original_app_path = strdup(menu_info->app_path);
	ptr = bundle_get_val(kb, AUL_K_PACKAGETYPE);
	if (ptr)
		menu_info->pkg_type = strdup(ptr);
	ptr = bundle_get_val(kb, AUL_K_APP_TYPE);
	if (ptr)
		menu_info->app_type = strdup(ptr);
	ptr = bundle_get_val(kb, AUL_K_HWACC);
	if (ptr)
		menu_info->hwacc = strdup(ptr);
	ptr = bundle_get_val(kb, AUL_K_TASKMANAGE);
	if (ptr)
		menu_info->taskmanage = strdup(ptr);
	ptr = bundle_get_val(kb, AUL_K_PKGID);
	if (ptr)
		menu_info->pkgid = strdup(ptr);
	ptr = bundle_get_val(kb, AUL_K_COMP_TYPE);
	if (ptr)
		menu_info->comp_type = strdup(ptr);
	ptr = bundle_get_val(kb, AUL_K_INTERNAL_POOL);
	if (ptr)
		menu_info->internal_pool = strdup(ptr);
	ptr = bundle_get_val(kb, AUL_K_ROOT_PATH);
	if (ptr)
		menu_info->root_path = strdup(ptr);

	ptr = bundle_get_val(kb, AUL_K_LOADER_NAME);
	if (ptr)
		menu_info->loader_name = strdup(ptr);

	ptr = bundle_get_val(kb, AUL_K_IS_GLOBAL);
	if (ptr && strcmp(ptr, "true") == 0)
		menu_info->global = true;
	else
		menu_info->global = false;

	if (!_appinfo_get_app_path(menu_info)) {
		_appinfo_free(menu_info);
		return NULL;
	}

	return menu_info;
}

char *_appinfo_get_app_path(appinfo_t *menu_info)
{
	int i = 0;
	int path_len = -1;
	char *tmp_app_path;

	if (!menu_info || menu_info->app_path == NULL)
		return NULL;

	while (menu_info->app_path[i] != 0) {
		if (menu_info->app_path[i] == ' '
		    || menu_info->app_path[i] == '\t') {
			path_len = i;
			break;
		}
		i++;
	}

	if (path_len == 0) {
		free(menu_info->app_path);
		menu_info->app_path = NULL;
	} else if (path_len > 0) {
		tmp_app_path = malloc(sizeof(char) * (path_len + 1));
		if (tmp_app_path == NULL)
			return NULL;
		snprintf(tmp_app_path, path_len + 1, "%s", menu_info->app_path);
		free(menu_info->app_path);
		menu_info->app_path = tmp_app_path;
	}

	return menu_info->app_path;
}

void _appinfo_free(appinfo_t *menu_info)
{
	if (menu_info == NULL)
		return;

	if (menu_info->appid != NULL)
		free(menu_info->appid);
	if (menu_info->app_path != NULL)
		free(menu_info->app_path);
	if (menu_info->original_app_path != NULL)
		free(menu_info->original_app_path);
	if (menu_info->pkg_type != NULL)
		free(menu_info->pkg_type);
	if (menu_info->app_type != NULL)
		free(menu_info->app_type);
	if (menu_info->hwacc != NULL)
		free(menu_info->hwacc);
	if (menu_info->taskmanage != NULL)
		free(menu_info->taskmanage);
	if (menu_info->pkgid != NULL)
		free(menu_info->pkgid);
	if (menu_info->comp_type != NULL)
		free(menu_info->comp_type);
	if (menu_info->internal_pool != NULL)
		free(menu_info->internal_pool);
	if (menu_info->root_path != NULL)
		free(menu_info->root_path);
	if (menu_info->loader_name != NULL)
		free(menu_info->loader_name);

	free(menu_info);
}

static bool __validate_bundle_key(const char *key)
{
	if (!key)
		return false;

	if (!strncmp(key, "__AUL_", strlen("__AUL_")))
		return false;

	if (!strncmp(key, "__APP_SVC_", strlen("__APP_SVC_")))
		return false;

	return true;
}

void _modify_bundle(bundle *kb, int caller_pid, appinfo_t *menu_info, int cmd)
{
	char *ptr;
	char exe[MAX_PATH_LEN];
	int flag;
	char key[256];
	char value[256];

	bundle_del(kb, AUL_K_APPID);
	bundle_del(kb, AUL_K_EXEC);
	bundle_del(kb, AUL_K_APP_TYPE);
	bundle_del(kb, AUL_K_PACKAGETYPE);
	bundle_del(kb, AUL_K_TASKMANAGE);
	bundle_del(kb, AUL_K_PKGID);
	bundle_del(kb, AUL_K_COMP_TYPE);

	/* Parse app_path to retrieve default bundle*/
	if (cmd == PAD_CMD_LAUNCH) {
		ptr = menu_info->original_app_path;
		flag = __parse_app_path(ptr, exe, sizeof(exe));
		if (flag > 0) {
			ptr += flag;
			SECURE_LOGD("parsing app_path: EXEC - %s", exe);

			do {
				flag = __parse_app_path(ptr, key, sizeof(key));
				if (flag <= 0)
					break;
				ptr += flag;

				flag = __parse_app_path(ptr, value,
						sizeof(value));
				if (flag < 0)
					break;
				ptr += flag;

				/*bundle_del(kb, key);*/
				if (__validate_bundle_key(key))
					bundle_add(kb, key, value);
			} while (flag > 0);
		} else if (flag == 0)
			_D("parsing app_path: No arguments");
		else
			_D("parsing app_path: Invalid argument");
	}
}

int _connect_to_launchpad(int type, int id)
{
	char path[PATH_MAX];
	int client_pid;
	int send_ret;
	int fd;

	_D("[launchpad] enter, type: %d", type);

	snprintf(path, sizeof(path), "%s/daemons/%d/%s%d-%d",
			SOCKET_PATH, getuid(), LAUNCHPAD_LOADER_SOCKET_NAME,
			type, id);
	fd = __create_client_socket(path);
	if (fd < 0)
		return -1;

	client_pid = getpid();
	send_ret = send(fd, &client_pid, sizeof(client_pid), MSG_NOSIGNAL);
	_D("send(%d) : %d", client_pid, send_ret);
	if (send_ret == -1) {
		_E("send error");
		close(fd);
		return -1;
	}

	SECURE_LOGD("[launchpad] done, connect fd: %d", fd);
	return fd;
}

#ifdef TIZEN_FEATURE_SET_PERSONALITY_32
static void __set_execution_domain(void)
{
	char err_buf[1024];
	int res;

	res = personality(PER_LINUX32);
	if (res < 0) {
		_E("personality() failed,  error: %d (%s)",
				errno,
				strerror_r(errno, err_buf, sizeof(err_buf)));
	}
}
#endif /* TIZEN_FEATURE_SET_PERSONALITY_32 */

void _set_env(appinfo_t *menu_info, bundle *kb)
{
	const char *str;
	char buf[MAX_LOCAL_BUFSZ];

	str = bundle_get_val(kb, AUL_K_STARTTIME);
	if (str != NULL)
		setenv("APP_START_TIME", str, 1);

	if (menu_info->hwacc != NULL)
		setenv("HWACC", menu_info->hwacc, 1);
	if (menu_info->taskmanage != NULL)
		setenv("TASKMANAGE", menu_info->taskmanage, 1);
	if (menu_info->root_path != NULL)
		setenv("AUL_ROOT_PATH", menu_info->root_path, 1);
	if (menu_info->appid != NULL)
		setenv("AUL_APPID", menu_info->appid, 1);
	if (menu_info->pkgid != NULL)
		setenv("AUL_PKGID", menu_info->pkgid, 1);
	if (menu_info->app_type != NULL)
		setenv("RUNTIME_TYPE", menu_info->app_type, 1);

	str = bundle_get_val(kb, AUL_K_WAYLAND_DISPLAY);
	if (str != NULL)
		setenv("WAYLAND_DISPLAY", str, 1);

	str = bundle_get_val(kb, AUL_K_WAYLAND_WORKING_DIR);
	if (str != NULL)
		setenv("XDG_RUNTIME_DIR", str, 1);

	str = bundle_get_val(kb, AUL_K_API_VERSION);
	if (str != NULL)
		setenv("TIZEN_API_VERSION", str, 1);

	snprintf(buf, sizeof(buf), "%d", getpid());
	setenv("AUL_PID", buf, 1);

#ifdef TIZEN_FEATURE_SET_PERSONALITY_32
	__set_execution_domain();
#endif /* TIZEN_FEATURE_SET_PERSONALITY_32 */
}

char **_create_argc_argv(bundle *kb, int *margc)
{
	char **argv;
	int argc;

	argc = bundle_export_to_argv(kb, &argv);

	*margc = argc;
	return argv;
}

char *_get_libdir(const char *path)
{
	char *path_dup;
	char buf[PATH_MAX];
	char *ptr;

	if (path == NULL)
		return NULL;
	path_dup = strdup(path);
	if (path_dup == NULL)
		return NULL;
	ptr = strrchr(path_dup, '/');
	if (ptr == NULL) {
		free(path_dup);
		return NULL;
	}

	*ptr = '\0';

	snprintf(buf, sizeof(buf), "%s/../lib/", path_dup);
	free(path_dup);

	if (access(buf, F_OK) == -1)
		return NULL;

	return strdup(buf);
}

int _proc_get_attr_by_pid(int pid, char *buf, int size)
{
	char path[PATH_MAX] = { 0, };
	int ret;

	snprintf(path, sizeof(path), "/proc/%d/attr/current", pid);
	ret = __read_proc(path, buf, size);
	if (ret <= 0)
		return -1;

	return 0;
}

static int __delete_dir(const char *path)
{
	DIR *dp;
	struct dirent *dentry = NULL;
	char buf[PATH_MAX];
	struct stat statbuf;
	int ret;

	if (path == NULL)
		return -1;

	dp = opendir(path);
	if (dp == NULL)
		return -1;

	while ((dentry = readdir(dp)) != NULL) {
		if (!strcmp(dentry->d_name, ".") || !strcmp(dentry->d_name, ".."))
			continue;

		snprintf(buf, sizeof(buf), "%s/%s", path, dentry->d_name);
		ret = stat(buf, &statbuf);
		if (ret == 0) {
			if (S_ISDIR(statbuf.st_mode))
				__delete_dir(buf);
			else
				unlink(buf);
		}
	}

	rmdir(path);
	closedir(dp);

	return 0;
}

int _delete_sock_path(int pid, uid_t uid)
{
	char path[PATH_MAX];
	int r;

	snprintf(path, sizeof(path), "/run/aul/apps/%d/%d", uid, pid);
	r = access(path, F_OK);
	if (r != 0) {
		if (errno != ENOENT)
			_E("Failed to access %s. errno(%d)", path, errno);
	} else {
		__delete_dir(path);
	}

	if (access(path, F_OK) == 0) {
		_E("Failed to delete %s", path);
		return -1;
	}

	return 0;
}

int _close_all_fds(void)
{
	DIR *dp;
	struct dirent *dentry = NULL;
	int fd;
	int max_fd;
	int aul_fd = -1;
	const char *aul_listen_fd;

	aul_listen_fd = getenv("AUL_LISTEN_FD");
	if (aul_listen_fd)
		aul_fd = atoi(aul_listen_fd);

	dp = opendir("/proc/self/fd");
	if (dp == NULL) {
		/* fallback */
		max_fd = sysconf(_SC_OPEN_MAX);
		for (fd = 3; fd < max_fd; fd++) {
			if (fd != aul_fd)
				close(fd);
		}

		return 0;
	}

	while ((dentry = readdir(dp)) != NULL) {
		if (!isdigit(dentry->d_name[0]))
			continue;

		fd = atoi(dentry->d_name);
		if (fd < 3)
			continue;

		if (fd == dirfd(dp) || fd == aul_fd)
			continue;

		close(fd);
	}
	closedir(dp);

	return 0;
}

int _setup_stdio(const char *ident)
{
	int fd;

	/* stdin */
	fd = open(PATH_DEV_NULL, O_RDONLY | O_NOCTTY);
	if (fd < 0) {
		_W("Failed to open /dev/null - err(%d)", errno);
		return -1;
	}
	if (dup2(fd, STDIN_FILENO) < 0) {
		_W("Failed to duplicate fd - oldfd(%d), newfd(%d)",
				fd, STDIN_FILENO);
	}
	close(fd);

	/* stdout */
	fd = sd_journal_stream_fd(ident, LOG_INFO, false);
	if (fd < 0) {
		_W("Failed to connect journal socket - err(%d)", errno);
		fd = open(PATH_DEV_NULL, O_WRONLY | O_NOCTTY);
		if (fd < 0) {
			_W("Failed to open /dev/null - err(%d)", errno);
			return -1;
		}
	}
	if (dup2(fd, STDOUT_FILENO) < 0) {
		_W("Failed to duplicate fd - oldfd(%d), newfd(%d)",
				fd, STDOUT_FILENO);
	}
	close(fd);

	/* stderr */
	fd = sd_journal_stream_fd(ident, LOG_INFO, false);
	if (fd < 0) {
		_W("Failed to connect journal socket - err(%d)", errno);
		fd = open(PATH_DEV_NULL, O_WRONLY | O_NOCTTY);
		if (fd < 0) {
			_W("Failed to open /dev/null - err(%d)", errno);
			return -1;
		}
	}
	if (dup2(fd, STDERR_FILENO) < 0) {
		_W("Failed to duplicate fd - oldfd(%d), newfd(%d)",
				fd, STDERR_FILENO);
	}
	close(fd);

	return 0;
}

int _set_priority(int prio)
{
	int r;

	r = setpriority(PRIO_PGRP, 0, prio);
	if (r < 0) {
		SECURE_LOGE("Failed to set process(%d) priority(%d) - err(%d)",
				getpid(), prio, errno);
	}

	return r;
}

static int __dbus_send_message(DBusMessage *callback(void *), void *user_data,
		int timeout, int *result)
{
	DBusError err;
	DBusConnection *conn;
	DBusMessage *msg = NULL;
	DBusMessageIter args;
	DBusPendingCall *pending = NULL;
	dbus_bool_t r;
	dbus_int32_t i;
	int ret = 0;

	dbus_error_init(&err);
	conn = dbus_bus_get_private(DBUS_BUS_SYSTEM, &err);
	if (!conn) {
		_E("Failed to connect to D-Bus Daemon");
		ret = -1;
		goto end;
	}

	msg = callback(user_data);
	if (!msg) {
		_E("Message is nullptr");
		goto end;
	}

	r = dbus_connection_send_with_reply(conn, msg, &pending, timeout);
	if (!r || !pending) {
		_E("Failed to send message");
		ret = -1;
		goto end;
	}

	dbus_connection_flush(conn);
	dbus_message_unref(msg);

	dbus_pending_call_block(pending);
	msg = dbus_pending_call_steal_reply(pending);
	if (!msg) {
		_E("Failed to get reply message");
		ret = -1;
		goto end;
	}

	dbus_pending_call_unref(pending);

	if (!dbus_message_iter_init(msg, &args)) {
		_E("Message has no arguments!");
		ret = -1;
	} else if (dbus_message_iter_get_arg_type(&args) != DBUS_TYPE_INT32) {
		_E("Argument is not integer! type(%d)",
				dbus_message_iter_get_arg_type(&args));
		ret = -1;
	} else {
		dbus_message_iter_get_basic(&args, &i);
		*result = i;
	}

	_D("Result: %d", *result);
end:
	if (msg)
		dbus_message_unref(msg);

	if (conn)
		dbus_connection_close(conn);

	if (dbus_error_is_set(&err)) {
		_E("D-Bus error(%s)", err.message);
		dbus_error_free(&err);
	}

	return ret;
}

static DBusMessage *__create_tep_mount_message(void *user_data)
{
	DBusMessage *message;
	DBusMessageIter args;
	dbus_bool_t r;
	const char *tep_path = (const char *)user_data;

	message = dbus_message_new_method_call(TEP_BUS_NAME,
			TEP_OBJECT_PATH,
			TEP_INTERFACE_NAME,
			TEP_IS_MOUNTED_METHOD);
	if (!message) {
		_E("Message is nullptr");
		return NULL;
	}

	dbus_message_iter_init_append(message, &args);
	r = dbus_message_iter_append_basic(&args, DBUS_TYPE_STRING, &tep_path);
	if (!r) {
		_E("Out of memory");
		dbus_message_unref(message);
		return NULL;
	}

	return message;
}

static int __is_tep_mount_done(const char *tep_path)
{
	int result = -1;
	int ret;

	ret = __dbus_send_message(__create_tep_mount_message,
			(void *)tep_path, 500, &result);
	if (ret != 0)
		_E("Failed to send message");

	return result;
}

static int __check_tep_mount(const char *tep_path)
{
	int r;
	int cnt = 0;

	if (!tep_path)
		return 0;

	while (cnt < MAX_TEP_IS_MOUNT_RETRY_CNT) {
		r = __is_tep_mount_done(tep_path);
		if (r == 1)
			break;
		usleep(50 * 1000);
		cnt++;
	}

	if (r != 1) {
		_E("Not able to mount within 5 sec: %s", tep_path);
		return -1;
	}

	return 0;
}

int _wait_tep_mount(bundle *b)
{
	int r;
	int i;
	int len = 0;
	const char **tep_paths;

	tep_paths = bundle_get_str_array(b, AUL_K_TEP_PATH, &len);
	if (!tep_paths)
		return 0;

	for (i = 0; i < len; ++i) {
		r = __check_tep_mount(tep_paths[i]);
		if (r < 0)
			return -1;
	}

	_I("Mount has been done.");
	return 0;
}

static int __create_app_socket(int pid, uid_t uid)
{
	int fd;
	char path[PATH_MAX];
	struct sockaddr_un addr = {
		.sun_family = AF_UNIX,
	};

	fd = socket(AF_UNIX, SOCK_STREAM | SOCK_NONBLOCK, 0);
	if (fd < 0) {
		if (errno == EINVAL) {
			fd = socket(AF_UNIX, SOCK_STREAM, 0);
			if (fd < 0) {
				_E("Second chance - socket create errno(%d)",
						errno);
				return -1;
			}
		} else {
			_E("Failed to create socket. errno(%d)", errno);
			return -1;
		}
	}

	snprintf(path, sizeof(path), "/run/aul/apps/%u/%d", uid, pid);
	if (mkdir(path, 0700) != 0) {
		if (errno == EEXIST) {
			if (access(path, R_OK) != 0) {
				_E("Failed to access %s. errno(%d)",
						path, errno);
				close(fd);
				return -errno;
			}
		} else {
			_E("Failed to access %s. errno(%d)", path, errno);
			close(fd);
			return -1;
		}
	}

	snprintf(addr.sun_path, sizeof(addr.sun_path),
			"/run/aul/apps/%u/%d/.app-sock", uid, pid);
	unlink(addr.sun_path);

	if (bind(fd, (struct sockaddr *)&addr, sizeof(addr)) < 0) {
		_E("Failed to bind socket(%d), errno(%d)", fd, errno);
		close(fd);
		return -1;
	}

	if (_set_sock_option(fd, 0) < 0) {
		_E("Failed to set sock option");
		close(fd);
		return -1;
	}

	if (listen(fd, 128) < 0) {
		_E("Failed to listen %d, errno(%d)", fd, errno);
		close(fd);
		return -1;
	}

	return fd;
}

int _prepare_app_socket(void)
{
	int fd;
	char buf[12];

	fd = __create_app_socket(getpid(), getuid());
	if (fd < 0)
		return fd;

	snprintf(buf, sizeof(buf), "%d", fd);
	setenv("AUL_LISTEN_FD", buf, 1);

	return 0;
}

struct package_info_s {
	const char *id;
	uid_t uid;
};

static DBusMessage *__create_app2sd_message(void *user_data)
{
	struct package_info_s *pkg_info = (struct package_info_s *)user_data;
	DBusMessage *message;
	DBusMessageIter args;
	dbus_bool_t r;

	_D("[__APP2SD__] package(%s), uid(%u)", pkg_info->id, pkg_info->uid);

	message = dbus_message_new_method_call(APP2SD_BUS_NAME,
			APP2SD_OBJECT_PATH,
			APP2SD_INTERFACE_NAME,
			APP2SD_ONDEMANDSETUPINIT_METHOD);
	if (!message) {
		_E("Message is nullptr");
		return NULL;
	}

	dbus_message_iter_init_append(message, &args);
	r = dbus_message_iter_append_basic(&args, DBUS_TYPE_STRING,
			&pkg_info->id);
	if (!r) {
		_E("Failed to append pkgid");
		dbus_message_unref(message);
		return NULL;
	}

	r = dbus_message_iter_append_basic(&args, DBUS_TYPE_INT32,
			&pkg_info->uid);
	if (!r) {
		_E("Failed to append uid");
		dbus_message_unref(message);
		return NULL;
	}

	return message;
}

int _enable_external_pkg(bundle *b, const char *pkgid, uid_t pkg_uid)
{
	const char *installed_storage;
	int retry_cnt = 0;
	int result = -1;
	int ret;
	struct package_info_s pkg_info = {
		.id = pkgid,
		.uid = pkg_uid
	};

	installed_storage = bundle_get_val(b, AUL_K_INSTALLED_STORAGE);
	if (!installed_storage || strcmp(installed_storage, "external") != 0)
		return 0;

	do {
		ret = __dbus_send_message(__create_app2sd_message,
				&pkg_info, 500, &result);
		if (ret != 0) {
			_E("Failed to send message");
			retry_cnt++;
			usleep(APP2SD_WAIT_USEC);
			continue;
		}

		break;
	} while (retry_cnt <= APP2SD_RETRY_MAX);

	_D("[__APP2SD__] result(%d)", result);

	return result;
}

int _verify_proc_caps(void)
{
	cap_t cap_d;
	cap_flag_value_t eff_state;
	cap_flag_value_t inh_state;
	cap_value_t value = CAP_SETGID;
	int r;

	cap_d = cap_get_proc();
	if (!cap_d) {
		_E("Failed to get cap from proc. pid(%d)", getpid());
		return -1;
	}

	r = cap_get_flag(cap_d, value, CAP_INHERITABLE, &inh_state);
	if (r != 0) {
		_E("Failed to get cap inh - errno(%d)", errno);
		cap_free(cap_d);
		return -1;
	}

	r = cap_get_flag(cap_d, value, CAP_EFFECTIVE, &eff_state);
	if (r != 0) {
		_E("Failed to get cap eff - errno(%d)", errno);
		cap_free(cap_d);
		return -1;
	}

	if ((inh_state != CAP_SET) || (eff_state != CAP_SET)) {
		_E("The process(%d) doesn't have %d cap",
				getpid(), value);
		cap_free(cap_d);
		return -1;
	}
	cap_free(cap_d);

	return 0;
}

int _prepare_id_file(void)
{
	char path[PATH_MAX];
	int fd;

	snprintf(path, sizeof(path), "/run/aul/apps/%u/%d/%s",
			getuid(), getpid(), getenv("AUL_APPID"));
	fd = open(path, O_CREAT | O_WRONLY | O_TRUNC, 0600);
	if (fd < 0) {
		_E("Failed to create %s. errno(%d)", path, errno);
		return -1;
	}
	close(fd);

	return 0;
}

static int __send_raw(int fd, int cmd, unsigned char *data, int datalen)
{
	app_pkt_t *pkt;
	int ret;

	pkt = (app_pkt_t *)calloc(1, sizeof(app_pkt_t) + datalen);
	if (!pkt) {
		_E("Out of memory");
		return -ENOMEM;
	}

	pkt->cmd = cmd;
	pkt->len = datalen;
	if (data)
		memcpy(pkt->data, data, pkt->len);

	ret = _send_pkt_raw(fd, pkt);
	if (ret < 0) {
		free(pkt);
		return -1;
	}

	free(pkt);

	return 0;
}

int _send_message_to_logger(const char *tag, const char *format, ...)
{
	char fmt[PATH_MAX];
	char buf[PATH_MAX];
	va_list ap;
	int ret;
	int fd;

	snprintf(buf, sizeof(buf), "%s/daemons/%u/.launchpad-logger-sock",
			SOCKET_PATH, getuid());
	fd = __create_client_socket(buf);
	if (fd < 0)
		return -1;

	va_start(ap, format);
	vsnprintf(fmt, sizeof(fmt), format, ap);
	va_end(ap);

	snprintf(buf, sizeof(buf), "%s: %s", tag, fmt);
	ret = __send_raw(fd, 0, (unsigned char *)buf, strlen(buf) + 1);
	if (ret < 0) {
		close(fd);
		return -1;
	}

	close(fd);

	return 0;
}

void _print_hwc_log(const char *format, ...)
{
	char buf[1024];
	va_list ap;

	va_start(ap, format);
	vsnprintf(buf, sizeof(buf), format, ap);
	va_end(ap);

	prctl(PR_TASK_PERF_USER_TRACE, buf, strlen(buf));
}
