/*
 * Copyright (c) 2019 Samsung Electronics Co., Ltd All Rights Reserved
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#define _GNU_SOURCE
#include <stdio.h>
#include <malloc.h>
#include <stdlib.h>
#include <dirent.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <limits.h>
#include <iniparser.h>
#include <vconf.h>

#include "launchpad_common.h"
#include "launchpad_config.h"

#define PATH_LAUNCHPAD_CONF             "/usr/share/aul/launchpad.conf"
#define TAG_MEMORY_STATUS               "MemoryStatus"
#define KEY_MEMORY_STATUS_LOW_KEY       "LowKey"
#define KEY_MEMORY_STATUS_LOW_VALUE     "LowValue"
#define KEY_MEMORY_STATUS_NORMAL_KEY    "NormalKey"
#define KEY_MEMORY_STATUS_NORMAL_VALUE  "NormalValue"

struct memory_status_s {
	char *low_key;
	int low_value;
	char *normal_key;
	int normal_value;
};

static struct memory_status_s __memory_status;

const char *_config_get_string_value(config_type_e type)
{
	const char *value;

	switch (type) {
	case CONFIG_TYPE_MEMORY_STATUS_LOW_KEY:
		value = __memory_status.low_key;
		break;
	case CONFIG_TYPE_MEMORY_STATUS_NORMAL_KEY:
		value = __memory_status.normal_key;
		break;
	default:
		_E("Unknown type");
		value = NULL;
		break;
	}

	return value;
}

int _config_get_int_value(config_type_e type)
{
	int value;

	switch (type) {
	case CONFIG_TYPE_MEMORY_STATUS_LOW_VALUE:
		value = __memory_status.low_value;
		break;
	case CONFIG_TYPE_MEMORY_STATUS_NORMAL_VALUE:
		value = __memory_status.normal_value;
		break;
	default:
		_E("Unknown type");
		value = INT_MIN;
	}

	return value;
}

static char *__get_string_value(dictionary *d, const char *tag,
		const char *key)
{
	char buf[128];

	snprintf(buf, sizeof(buf), "%s:%s", tag, key);
	return iniparser_getstring(d, buf, NULL);
}

static int __get_int_value(dictionary *d, const char *tag,
		const char *key)
{
	char buf[128];

	snprintf(buf, sizeof(buf), "%s:%s", tag, key);
	return iniparser_getint(d, buf, INT_MIN);
}

int _config_init(void)
{
	dictionary *d;
	char *val;
	char *str;
	int ret;

	_D("config init");

	__memory_status.low_key = strdup(VCONFKEY_SYSMAN_LOW_MEMORY);
	__memory_status.low_value = VCONFKEY_SYSMAN_LOW_MEMORY_SOFT_WARNING;
	__memory_status.normal_key = strdup(VCONFKEY_SYSMAN_LOW_MEMORY);
	__memory_status.normal_value = VCONFKEY_SYSMAN_LOW_MEMORY_NORMAL;

	ret = access(PATH_LAUNCHPAD_CONF, F_OK);
	if (ret != 0) {
		_W("The file doesn't exist. errno(%d)", errno);
		return 0;
	}

	d = iniparser_load(PATH_LAUNCHPAD_CONF);
	if (!d) {
		_E("Failed to load the config file");
		return -1;
	}

	str = __get_string_value(d, TAG_MEMORY_STATUS,
			KEY_MEMORY_STATUS_LOW_KEY);
	if (str) {
		val = strdup(str);
		if (val) {
			free(__memory_status.low_key);
			__memory_status.low_key = val;
		}
	}

	ret = __get_int_value(d, TAG_MEMORY_STATUS,
			KEY_MEMORY_STATUS_LOW_VALUE);
	if (ret != INT_MIN)
		__memory_status.low_value = ret;

	str = __get_string_value(d, TAG_MEMORY_STATUS,
			KEY_MEMORY_STATUS_NORMAL_KEY);
	if (str) {
		val = strdup(str);
		if (val) {
			free(__memory_status.normal_key);
			__memory_status.normal_key = val;
		}
	}

	ret = __get_int_value(d, TAG_MEMORY_STATUS,
			KEY_MEMORY_STATUS_NORMAL_VALUE);
	if (ret != INT_MIN)
		__memory_status.normal_value = ret;

	_W("Memory Status Low(%s:%d), Normal(%s:%d)",
			__memory_status.low_key,
			__memory_status.low_value,
			__memory_status.normal_key,
			__memory_status.normal_value);

	iniparser_freedict(d);

	return 0;
}

void _config_fini(void)
{
	_D("config fini");

	free(__memory_status.normal_key);
	free(__memory_status.low_key);
}
