/*
 * Copyright (c) 2016 Samsung Electronics Co., Ltd All Rights Reserved
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an AS IS BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#define _GNU_SOURCE
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <linux/limits.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <bundle_internal.h>

#include "debugger_info.h"
#include "key.h"
#include "launchpad_common.h"
#include "launchpad_debug.h"
#include "launchpad_types.h"

#define DEBUGGER_INFO_PATH "/usr/share/aul"

static int debug_initialized;
static GList *debugger_info_list;
static debugger_info_h debugger_info;
static GList *debug_argv_list;
static GList *extra_argv_list;

int _debug_create_extra_argv(int *argc, char ***argv)
{
	int new_argc;
	char **new_argv;
	const char *extra_argv;
	GList *iter;
	int i;

	if (argc == NULL || argv == NULL) {
		_E("[DEBUG] Invalid parameter");
		return -1;
	}

	if (debugger_info == NULL)
		return 0;

	new_argc = g_list_length(extra_argv_list);
	if (new_argc == 0)
		return 0;

	new_argv = (char **)calloc(new_argc, sizeof(char *));
	if (new_argv == NULL) {
		_E("out of memory");
		return -1;
	}

	i = LOADER_ARG_PATH;
	iter = g_list_first(extra_argv_list);
	while (iter) {
		extra_argv = (const char *)iter->data;
		if (extra_argv)
			new_argv[i++] = strdup(extra_argv);

		iter = g_list_next(iter);
	}

	*argc = new_argc;
	*argv = new_argv;
	_D("[DEBUG] argc: %d, i: %d", *argc, i);

	return 0;
}

int _debug_create_argv(int *argc, char ***argv, bool *attach)
{
	int new_argc = 0;
	char **new_argv;
	const char *exe;
	const char *debug_argv;
	const char *attach_str;
	GList *iter;
	GList *list;
	int i;

	if (argc == NULL || argv == NULL || attach == NULL) {
		_E("[DEBUG] Invalid parameter");
		return -1;
	}

	if (debugger_info == NULL)
		return 0;

	exe = _debugger_info_get_exe(debugger_info);
	if (exe == NULL)
		return 0;

	attach_str = _debugger_info_get_attach(debugger_info);
	if (attach_str && strcasecmp(attach_str, "true") == 0) {
		*attach = true;
		new_argc++;
	}

	list = _debugger_info_get_default_opt_list(debugger_info);
	new_argc += g_list_length(debug_argv_list) + g_list_length(list) + 1;
	new_argv = (char **)calloc(new_argc, sizeof(char *));
	if (new_argv == NULL) {
		_E("out of memory");
		return -1;
	}

	i = LOADER_ARG_PATH;
	new_argv[i++] = strdup(exe);

	iter = g_list_first(list);
	while (iter) {
		debug_argv = (const char *)iter->data;
		if (debug_argv)
			new_argv[i++] = strdup(debug_argv);

		iter = g_list_next(iter);
	}

	iter = g_list_first(debug_argv_list);
	while (iter) {
		debug_argv = (const char *)iter->data;
		if (debug_argv)
			new_argv[i++] = strdup(debug_argv);

		iter = g_list_next(iter);
	}

	*argc = new_argc;
	*argv = new_argv;
	_D("[DEBUG] argc: %d, argv[0]: %s",
			new_argc, new_argv[LOADER_ARG_PATH]);

	return 0;
}

void _debug_destroy_argv(int argc, char **argv)
{
	int i;

	if (argv == NULL)
		return;

	for (i = 0; i < argc; i++)
		free(argv[i]);
	free(argv);
}

int _debug_get_caller_pid(bundle *kb)
{
	const char *pid_str;
	int pid;

	pid_str = bundle_get_val(kb, AUL_K_ORG_CALLER_PID);
	if (pid_str == NULL)
		pid_str = bundle_get_val(kb, AUL_K_CALLER_PID);

	if (pid_str == NULL)
		return -1;

	pid = atoi(pid_str);
	if (pid <= 1)
		return -1;

	return pid;
}

static int __redirect_std_fds(bundle *kb)
{
	char path[PATH_MAX];
	char err_buf[1024];
	int fd;
	int caller_pid;

	if (kb == NULL) {
		_E("[DEBUG] Invalid parameter");
		return -1;
	}

	caller_pid = _debug_get_caller_pid(kb);
	if (caller_pid < 0) {
		_E("[DEBUG] Failed to get caller pid");
		return -1;
	}

	/* stdin */
	snprintf(path, sizeof(path), "/proc/%d/fd/0", caller_pid);
	fd = open(path, O_RDONLY);
	if (fd < 0) {
		_E("[DEBUG] Failed to open %s [%s]", path,
				strerror_r(errno, err_buf, sizeof(err_buf)));
		return -1;
	}
	dup2(fd, 0);
	close(fd);

	/* stdout */
	snprintf(path, sizeof(path), "/proc/%d/fd/1", caller_pid);
	fd = open(path, O_WRONLY);
	if (fd < 0) {
		_E("[DEBUG] Failed to open %s [%s]", path,
				strerror_r(errno, err_buf, sizeof(err_buf)));
		return -1;
	}
	dup2(fd, 1);
	close(fd);

	/* stderr */
	snprintf(path, sizeof(path), "/proc/%d/fd/2", caller_pid);
	fd = open(path, O_WRONLY);
	if (fd < 0) {
		_E("[DEBUG] Failed to open %s [%s]", path,
				strerror_r(errno, err_buf, sizeof(err_buf)));
		return -1;
	}
	dup2(fd, 2);
	close(fd);

	return 0;
}

static void __add_extra_argv(gpointer data, gpointer user_data)
{
	const char *key = (const char *)data;
	bundle *kb = (bundle *)user_data;
	const char *str;
	const char **str_arr = NULL;
	int len = 0;
	int i;

	if (key == NULL || kb == NULL)
		return;

	_D("[DEBUG] key: %s", key);
	if (bundle_get_type(kb, key) & BUNDLE_TYPE_ARRAY) {
		str_arr = bundle_get_str_array(kb, key, &len);
	} else {
		str = bundle_get_val(kb, key);
		if (str) {
			str_arr = &str;
			len = 1;
		}
	}

	for (i = 0; i < len; i++) {
		if (str_arr[i] == NULL)
			break;

		extra_argv_list = g_list_append(extra_argv_list,
				strdup(str_arr[i]));
	}

	if (str_arr)
		bundle_del(kb, key);
}

static void __add_debug_argv(gpointer data, gpointer user_data)
{
	const char *key = (const char *)data;
	bundle *kb = (bundle *)user_data;
	const char *str;
	const char **str_arr = NULL;
	int len = 0;
	int i;

	if (key == NULL || kb == NULL)
		return;

	_D("[DEBUG] key: %s", key);
	if (bundle_get_type(kb, key) & BUNDLE_TYPE_ARRAY) {
		str_arr = bundle_get_str_array(kb, key, &len);
	} else {
		str = bundle_get_val(kb, key);
		if (str) {
			str_arr = &str;
			len = 1;
		}
	}

	for (i = 0; i < len; i++) {
		if (str_arr[i] == NULL)
			break;

		debug_argv_list = g_list_append(debug_argv_list,
				strdup(str_arr[i]));
	}

	if (str_arr)
		bundle_del(kb, key);
}

static void __set_debug_env(gpointer data, gpointer user_data)
{
	const char *key = (const char *)data;
	bundle *kb = (bundle *)user_data;
	const char *str;
	const char **str_arr = NULL;
	int len = 0;
	int i;
	char buf[LINE_MAX] = {0,};

	if (key == NULL || kb == NULL)
		return;

	_D("[DEBUG] key: %s", key);
	if (bundle_get_type(kb, key) & BUNDLE_TYPE_ARRAY) {
		str_arr = bundle_get_str_array(kb, key, &len);
	} else {
		str = bundle_get_val(kb, key);
		if (str) {
			str_arr = &str;
			len = 1;
		}
	}

	if (str_arr == NULL)
		return;

	strncat(buf, str_arr[0], sizeof(buf) - strlen(buf) - 1);
	for (i = 1; i < len; i++) {
		if (str_arr[i] == NULL)
			break;

		strncat(buf, ",", sizeof(buf) - strlen(buf) - 1);
		strncat(buf, str_arr[i], sizeof(buf) - strlen(buf) - 1);
	}

	bundle_del(kb, key);
	_D("[DEBUG] name: %s, value: %s", key, buf);
	setenv(key, buf, 1);
}

static void __remove_file(gpointer data, gpointer user_data)
{
	const char *file = (const char *)data;

	if (file == NULL)
		return;

	_D("[DEBUG] file: %s", file);
	if (access(file, F_OK) == 0) {
		if (remove(file) != 0)
			_W("[DEBUG] Failed to remove %s", file);
	}
}

void _debug_prepare_debugger(bundle *kb)
{
	const char *debugger;
	GList *list;
	int ret;

	if (kb == NULL)
		return;

	debugger = bundle_get_val(kb, AUL_K_SDK);
	if (debugger == NULL)
		return;

	ret = __redirect_std_fds(kb);
	if (ret < 0)
		_E("[DEBUG] Failed to redirect standard fds");

	_D("[DEBUG] debugger: %s", debugger);
	debugger_info = _debugger_info_find(debugger_info_list, debugger);
	if (debugger_info == NULL)
		return;

	list = _debugger_info_get_unlink_list(debugger_info);
	g_list_foreach(list, __remove_file, NULL);

	list = _debugger_info_get_extra_env_list(debugger_info);
	g_list_foreach(list, __set_debug_env, kb);

	list = _debugger_info_get_extra_key_list(debugger_info);
	g_list_foreach(list, __add_debug_argv, kb);

	list = _debugger_info_get_last_extra_key_list(debugger_info);
	g_list_foreach(list, __add_extra_argv, kb);
}

int _debug_init(void)
{
	if (debug_initialized)
		return 0;

	debugger_info_list = _debugger_info_load(DEBUGGER_INFO_PATH);
	if (debugger_info_list == NULL)
		return -1;

	debug_initialized = 1;

	return 0;
}

void _debug_fini(void)
{
	if (!debug_initialized)
		return;

	_debugger_info_unload(debugger_info_list);
}
