/*
 * Copyright (c) 2019 Samsung Electronics Co., Ltd All Rights Reserved
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#define _GNU_SOURCE
#include <dlfcn.h>
#include <errno.h>
#include <linux/limits.h>
#include <malloc.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/prctl.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <sys/un.h>
#include <sys/wait.h>
#include <unistd.h>

#include <systemd/sd-event.h>

#include "launchpad_hydra.h"
#include "launchpad_types.h"
#include "log_private.h"

#ifndef API
#define API __attribute__ ((visibility("default")))
#endif

typedef enum {
        LAUNCHPAD_TYPE_UNSUPPORTED = -1,
        LAUNCHPAD_TYPE_USER = 1,
        LAUNCHPAD_TYPE_DYNAMIC = 100,
        LAUNCHPAD_TYPE_MAX
} launchpad_type_e;

typedef struct hydra_context_s {
	sigset_t oldmask;
	sd_event *event;
	sd_event_source *fd_source;
	sd_event_source *sig_source;
	int signal_fd;
	int client_fd;
	int candidate_pid;
	int argc;
	char **argv;
	hydra_lifecycle_callback_s *hydra_callbacks;
	void *user_data;
} hydra_context_t;

static hydra_context_t __context;

static void __fini_hydra(void);

static int __init_signal(void)
{
	sigset_t mask;
	int sfd;

	_D("[__HYDRA__] init signal");
	sigemptyset(&mask);
	sigaddset(&mask, SIGCHLD);

	if (sigprocmask(SIG_BLOCK, &mask, &__context.oldmask) < 0)
		_E("sigprocmask() is failed. errno(%d)", errno);
	else
		_D("SIGCHLD blocked");

	sfd = signalfd(-1, &mask, SFD_NONBLOCK | SFD_CLOEXEC);
	if (sfd < 0) {
		_E("Failed to create SIGCHLD fd. errno(%d)", errno);
		return -1;
	}

	return sfd;
}

static void __fini_signal(void)
{
	_D("[__HYDRA__] finish signal");
	if (sigprocmask(SIG_SETMASK, &__context.oldmask, NULL) < 0)
		_E("sigprocmask() is failed. errno(%d)", errno);
	else
		_D("SIGCHLD unblocked");
}

static int (*__security_manager_prepare_app_candidate)(void);
static int __load_security_manager_func(void)
{
#define PATH_SECURITY_MANAGER_CLIENT "/usr/lib/libsecurity-manager-client.so.3"
	void *handle;

	handle = dlopen(PATH_SECURITY_MANAGER_CLIENT, RTLD_LAZY | RTLD_LOCAL);
	if (!handle) {
		_E("Failed to open security manager client library. error(%s)",
				dlerror());
		return -1;
	}

	__security_manager_prepare_app_candidate = dlsym(handle,
			"security_manager_prepare_app_candidate");
	if (!__security_manager_prepare_app_candidate) {
		_E("Failed to load security_manager_prepare_app_candidate()");
		return -1;
	}

	return 0;
}

static int __fork_process(int (*child_fn)(void *), void *arg)
{
	int ret;
	pid_t pid;

	pid = fork();
	if (pid < 0) {
		_E("[__HYDRA__] Failed to create a child process. errno(%d)",
				errno);
		return pid;
	}

	if (pid == 0) {
		__load_security_manager_func();

		if (__security_manager_prepare_app_candidate) {
			_W("security_manager_prepare_app_candidate ++");
			ret = __security_manager_prepare_app_candidate();
			_W("security_manager_prepare_app_candidate --");
			if (ret != 0) {
				_E("Failed to prepare app candidate process. error(%d)",
						ret);
				exit(1);
			}
		}

		ret = child_fn(arg);
		_E("Failed to exec child process. errno(%d)", errno);
		exit(ret);
	}

	return pid;
}

static int __run_loader(void *arg)
{
	int ret = 0;

	_D("[__HYDRA__] Run loader. pid(%d)", getpid());
	__fini_signal();

	/* Set new sesssion ID & new process group ID */
	setsid();

	launchpad_hydra_exit();
	__fini_hydra();

	if (__context.hydra_callbacks->fork) {
		ret = __context.hydra_callbacks->fork(__context.argc,
				__context.argv, __context.user_data);
	}

	return ret;
}

static int __handle_launch_event(sd_event_source *s,
		int fd, uint32_t revents, void *user_data)
{
	enum hydra_cmd cmd = -1;
	int len;

retry_recv:
	len = recv(fd, &cmd, sizeof(cmd), MSG_WAITALL);
	if (len < 0) {
		if (errno == EINTR)
			goto retry_recv;
	}

	if (len < sizeof(cmd)) {
		_E("[__HYDRA__] Failed to recv data. errno(%d)", errno);
		goto err;
	}

	if (cmd == LAUNCH_CANDIDATE) {
		_D("[__HYDRA__] Launch new child");
	} else {
		_E("[__HYDRA__] Unknown command(%d)", cmd);
		goto err;
	}

	__context.candidate_pid = __fork_process(__run_loader, NULL);
	_D("[__HYDRA__] candidate process(%d)", __context.candidate_pid);

	return 0;
err:
	__fini_hydra();
	exit(-1);
	return -1;
}

static int __process_sigchld(struct signalfd_siginfo *info)
{
	union sigval val;
	int status;
	pid_t child_pid;
	pid_t child_pgid;

	child_pgid = getpgid(info->ssi_pid);
	_D("[__HYDRA__] Dead pid(%d), pgid(%d), signo(%d), status(%d)",
			info->ssi_pid, child_pgid, info->ssi_signo,
			info->ssi_status);

	while ((child_pid = waitpid(-1, &status, WNOHANG)) > 0) {
		if (child_pid == child_pgid)
			killpg(child_pgid, SIGKILL);

		val.sival_int = child_pid;
		sigqueue(getppid(), SIGCHLD, val);
	}

	return 0;
}

static int __handle_sigchld_event(sd_event_source *s,
		int fd, uint32_t revents, void *user_data)
{
	struct signalfd_siginfo siginfo;
	ssize_t size;

	do {
		size = read(fd, &siginfo, sizeof(struct signalfd_siginfo));
		if (size == 0)
			break;

		if (size != sizeof(struct signalfd_siginfo))
			break;

		__process_sigchld(&siginfo);
	} while (size > 0);

	return 0;
}

static int __create_client_socket(const char *path)
{
	struct sockaddr_un addr = { 0, };
	int retry = CONNECT_RETRY_COUNT;
	int fd;

	fd = socket(AF_UNIX, SOCK_STREAM | SOCK_CLOEXEC, 0);
	if (fd < 0) {
		_E("Failed to create socket(%s). errno(%d)", path, errno);
		return -1;
	}

	addr.sun_family = AF_UNIX;
	snprintf(addr.sun_path, sizeof(addr.sun_path), "%s", path);
	while (connect(fd, (struct sockaddr *)&addr, sizeof(addr)) < 0) {
		if (errno != ETIMEDOUT || retry <= 0) {
			_E("Failed to connect socket(%s). errno(%d)",
					path, errno);
			close(fd);
			return -1;
		}

		usleep(CONNECT_RETRY_TIME);
		retry--;
		_W("Retry(%d) to connect %s", retry, path);
	}

	return fd;
}

int __connect_to_launchpad_hydra(int type, int id)
{
	char path[PATH_MAX];
	int fd;
	int send_ret;
	pid_t client_pid;


	_D("[hydra] enter, type: %d", type);

	snprintf(path, sizeof(path), "%s/daemons/%d/%s%d-%d",
			SOCKET_PATH, getuid(), HYDRA_LOADER_SOCKET_NAME,
			type, id);
	fd = __create_client_socket(path);
	if (fd < 0)
		return -1;

	client_pid = getpid();
	send_ret = send(fd, &client_pid, sizeof(client_pid), MSG_NOSIGNAL);
	_D("send(%d) : %d", client_pid, send_ret);
	if (send_ret == -1) {
		_E("send error");
		close(fd);
		return -1;
	}

	SECURE_LOGD("[hydra] done, connect fd: %d", fd);
	return fd;
}

static int __init_hydra(int loader_type, int loader_id)
{
	int ret;

	_D("[__HYDRA__] init hydra");
	__context.signal_fd = __init_signal();
	if (__context.signal_fd < 0) {
		_E("Failed to initialize signal");
		return -1;
	}

	__context.client_fd = __connect_to_launchpad_hydra(loader_type,
			loader_id);
	if (__context.client_fd < 0) {
		_E("Failed to connect to launchpad");
		return -1;
	}

	ret = sd_event_default(&__context.event);
	if (ret < 0) {
		_E("Failed to create default sd event. error(%d)", ret);
		return ret;
	}

	sd_event_add_io(__context.event, &__context.fd_source,
			__context.client_fd, EPOLLIN,
			__handle_launch_event, NULL);
	sd_event_add_io(__context.event, &__context.sig_source,
			__context.signal_fd, EPOLLIN,
			__handle_sigchld_event, NULL);

	return 0;
}

static void __fini_hydra(void)
{
	_D("[__HYDRA__] finish hydra");
	if (__context.sig_source)
		sd_event_source_unref(__context.sig_source);

	if (__context.fd_source)
		sd_event_source_unref(__context.fd_source);

	if (__context.event)
		sd_event_unref(__context.event);

	if (__context.client_fd > 0)
		close(__context.client_fd);

	if (__context.signal_fd > 0)
		close(__context.signal_fd);
}

static void __run_loop(void)
{
	int ret;

	ret = sd_event_loop(__context.event);
	if (ret < 0)
		_E("Failed to run sd event loop. error(%d)", ret);

	_D("[__HYDRA__] Run main loop");
}

static void __exit_loop(void)
{
	if (__context.event)
		sd_event_exit(__context.event, 0);

	_D("[__HYDRA__] Exit main loop");
}

API int launchpad_hydra_main(int argc, char **argv,
		hydra_lifecycle_callback_s *hydra_callbacks,
		void *user_data)
{
	int loader_type;
	int loader_id;
	int is_hydra;
	int ret = 0;

	if (argc < 4) {
		_E("[__HYDRA__] too few argument");
		return -EINVAL;
	}

	loader_type = argv[LOADER_ARG_TYPE][0] - '0';
	if (loader_type < 0 || loader_type >= LAUNCHPAD_TYPE_MAX) {
		_E("[__HYDRA__] Invalid argument. type(%d)", loader_type);
		return -EINVAL;
	}

	if (!hydra_callbacks) {
		_E("[__HYDRA__] Invalid argument");
		return -EINVAL;
	}

	is_hydra = argv[LOADER_ARG_HYDRA][0] - '0';
	_D("[__HYDRA__] mode: %d", is_hydra);
	if (!is_hydra) {
		_W("Run in non hydra mode");
		return -EINVAL;
	}

	argv[LOADER_ARG_HYDRA] = "0";
	loader_id = atoi(argv[LOADER_ARG_ID]);

	__context.argc = argc;
	__context.argv = argv;
	__context.hydra_callbacks = hydra_callbacks;
	__context.user_data = user_data;

	ret = __init_hydra(loader_type, loader_id);
	if (ret < 0) {
		_E("Failed to initialize hydra loader");
		return ret;
	}

	if (__context.hydra_callbacks->precreate)
		__context.hydra_callbacks->precreate(user_data);

	if (__context.hydra_callbacks->create)
		__context.hydra_callbacks->create(__context.user_data);

	__context.candidate_pid = __fork_process(__run_loader, NULL);
	_D("[__HYDRA__] candidate process(%d)", __context.candidate_pid);

	__run_loop();

	if (__context.hydra_callbacks->terminate)
		ret = __context.hydra_callbacks->terminate(__context.user_data);

	__fini_hydra();

	return ret;
}

API void launchpad_hydra_exit(void)
{
	__exit_loop();
}
